{***********************************************}
{File:      NCNamExp.PAS                        }
{Revision:  1.03 / 06.02.2000                   }
{Comment:   Component naming expert             }
{Copyright: (c) 1997-2000, Dmitry Arefiev       }
{Author:    Dmitry Arefiev, dmitrya@inthink.com }
{***********************************************}
{$I NCOciDef.inc}

Unit NCNamExp;

interface

uses Classes, Forms, Menus, Windows, ExptIntf, ToolIntf, DsgnIntf;

type
    TNCComponentNamingExpert = class(TIExpert)
    private
        FNamingHotKey, FOptHotKey: TShortCut;
        EditRenameControlsSepItem, EditRenameControlsItem,
        EditRenameOptionsItem: TIMenuItemIntf;
        procedure EditRenameControlsClick(Sender: TIMenuItemIntf);
        procedure EditRenameSetupClick(Sender: TIMenuItemIntf);
        procedure SaveToReg;
        procedure LoadFromReg;
        function BaseRegKey: String;
        procedure SetNamingHotKey(AValue: TShortCut);
        procedure SetOptHotKey(AValue: TShortCut);
    public
        constructor Create;
        destructor Destroy; override;
        function GetName: string; override;
        function GetStyle: TExpertStyle; override;
        function GetIDString: string; override;
        property NamingHotKey: TShortCut read FNamingHotKey write SetNamingHotKey;
        property OptHotKey: TShortCut read FOptHotKey write SetOptHotKey;
    end;

    TNCComponentNameProperty = class(TStringProperty)
    public
        function GetAttributes: TPropertyAttributes; override;
        function GetEditLimit: Integer; override;
        procedure Edit; override;
        destructor Destroy; override;
        procedure Initialize; override;
        procedure SetValue(const Value: string); override;
        function GetValue: string; override;
    end;

    procedure RegisterNameExp;

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

implementation

Uses NCCmpNmr, NCNamOpt, Dialogs, SysUtils, NCStrs, NCUtil, Registry;

var
    CompNamer: TNCCompNamer = nil;
    CompNameProp: TNCComponentNameProperty = nil;

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

function TNCComponentNamingExpert.GetName: string;
begin
    Result := 'Component Naming Expert';
end;

function TNCComponentNamingExpert.GetStyle: TExpertStyle;
begin
    Result := esAddIn;
end;

function TNCComponentNamingExpert.GetIDString: string;
begin
    Result := 'NC.CompNamExp';
end;

function TNCComponentNamingExpert.BaseRegKey: String;
begin
    Result := ToolServices.GetBaseRegistryKey + '\NC\CompNamExp';
end;

procedure TNCComponentNamingExpert.SaveToReg;
var
    reg: TRegistry;
begin
    reg := TRegistry.Create;
    try
        reg.OpenKey(BaseRegKey, True);
        CompNamer.SaveToReg(reg);
        reg.WriteInteger('NamingHotKey', NamingHotKey);
        reg.WriteInteger('OptionsHotKey', OptHotKey);
    finally
        reg.Free;
    end;
end;

procedure TNCComponentNamingExpert.LoadFromReg;
var
    reg: TRegistry;
begin
    reg := TRegistry.Create;
    try
        if not reg.OpenKey(BaseRegKey, False) then
            Exit;
        CompNamer.LoadFromReg(reg);
        NamingHotKey := reg.ReadInteger('NamingHotKey');
        OptHotKey := reg.ReadInteger('OptionsHotKey');
    finally
        reg.Free;
    end;
end;

procedure TNCComponentNamingExpert.SetNamingHotKey(AValue: TShortCut);
begin
    FNamingHotKey := AValue;
    EditRenameControlsItem.SetShortCut(AValue);
end;

procedure TNCComponentNamingExpert.SetOptHotKey(AValue: TShortCut);
begin
    FOptHotKey := AValue;
    EditRenameOptionsItem.SetShortCut(AValue);
end;

constructor TNCComponentNamingExpert.Create;
var
    MainMenu: TIMainMenuIntf;
    EditLockControlsMenu: TIMenuItemIntf;
    EditMenu: TIMenuItemIntf;
    MenuItems: TIMenuItemIntf;
begin
    inherited Create;
    try
        MainMenu := ToolServices.GetMainMenu;
        if MainMenu <> nil then
        try
            MenuItems := MainMenu.GetMenuItems;
            if MenuItems <> nil then
            try
                EditLockControlsMenu := MainMenu.FindMenuItem('EditLockControlsItem');
                if EditLockControlsMenu <> nil then
                try
                    EditMenu := EditLockControlsMenu.GetParent;
                    if EditMenu <> nil then
                    try
                        EditRenameControlsSepItem := EditMenu.InsertItem(EditLockControlsMenu.GetIndex + 1,
                            '-', 'NCEditRenameControlsSepItem', '', 0, 0, 0,
                            [mfVisible], nil);
                        EditRenameControlsItem := EditMenu.InsertItem(EditLockControlsMenu.GetIndex + 2,
                            SEditRenameControlsItem, 'NCEditRenameControlsItem', '', 0, 0, 0,
                            [mfVisible, mfEnabled], EditRenameControlsClick);
                        EditRenameOptionsItem := EditMenu.InsertItem(EditLockControlsMenu.GetIndex + 3,
                            SEditRenameOptionsItem, 'NCEditRenameOptionsItem', '', 0, 0, 0,
                            [mfVisible, mfEnabled], EditRenameSetupClick);
                    finally
                        EditMenu.Free;
                    end;
                finally
                    EditLockControlsMenu.Free;
                end;
            finally
                MenuItems.Free;
            end;
        finally
            MainMenu.Free;
        end;
        NamingHotKey := ShortCut(Word('M'), [ssCtrl]);
        OptHotKey := ShortCut(Word('M'), [ssAlt, ssCtrl]);
        LoadFromReg;
    except
        Application.HandleException(Self);
    end;
end;

destructor TNCComponentNamingExpert.Destroy;
begin
    EditRenameControlsSepItem.Free;
    EditRenameControlsItem.Free;
    EditRenameOptionsItem.Free;
    inherited Destroy;
end;

procedure TNCComponentNamingExpert.EditRenameControlsClick(Sender: TIMenuItemIntf);
begin
    try
        if CompNameProp <> nil then
            CompNameProp.Edit;
    except
        Application.HandleException(Self);
    end;
end;

procedure TNCComponentNamingExpert.EditRenameSetupClick(Sender: TIMenuItemIntf);
begin
    try
        NCCompNamerEdt(CompNamer, Self);
        SaveToReg;
    except
        Application.HandleException(Self);
    end;
end;

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

function TNCComponentNameProperty.GetAttributes: TPropertyAttributes;
begin
    Result := [paMultiSelect, paDialog];
end;

function TNCComponentNameProperty.GetEditLimit: Integer;
begin
    Result := 63;
end;

procedure TNCComponentNameProperty.Edit;
var
    i: Integer;
begin
    i := 0;
    while True do
    try
        CompNamer.Component := GetComponent(i) as TComponent;
        CompNamer.Rename;
        Inc(i);
    except
        Designer.Modified;
        Break;
    end;
end;

function TNCComponentNameProperty.GetValue: string;
begin
    if PropCount > 1 then
        Result := SCallExpertNow
    else
        Result := inherited GetValue;
end;

procedure TNCComponentNameProperty.SetValue(const Value: string);
begin
    if PropCount > 1 then
        GenError(SUseExpertForRename)
    else
        inherited SetValue(Value);
end;

procedure TNCComponentNameProperty.Initialize;
begin
    inherited Initialize;
    CompNameProp := Self;
end;

destructor TNCComponentNameProperty.Destroy;
begin
    CompNameProp := nil;
    inherited Destroy;
end;

// -----------------------------------------------------------------------------
// -----------------------------------------------------------------------------

procedure RegisterNameExp;
begin
    RegisterLibraryExpert(TNCComponentNamingExpert.Create);
    RegisterPropertyEditor(TypeInfo(TComponentName), TComponent, 'Name',
        TNCComponentNameProperty);
end;

initialization

    CompNameProp := nil;
    CompNamer := TNCCompNamer.Create(nil);

finalization

    CompNamer.Free;

end.
