{***********************************************}
{File:          NCQBF.PAS                       }
{Revision:      2.11 / 13.07.1999               }
{Comment:       Query By Form system            }
{Copyright:     (c) 1997-1999, Nica-Com Ltd.    }
{E-Mail:        diman@ncom.nnov.su              }
{***********************************************}

unit NCQBF;

interface

Uses SysUtils, DB, DBTables, Classes, NCMemTbl, NCCompat, NCDBUtil;

{
==================================================================
State diagram:
==================================================================
                     FROM
    TO      qsInactive  qsEdit   qsActive
qsInactive      -       Cancel    Cancel
qsEdit        Edit        -        Edit
qsActive        X        Post       -

==================================================================
Language shema:
==================================================================
EQ = '=' | '>' | '<' | '>=' | '<=' | '<>'
mask = <chars>: '%' or '_' in <chars>
value = <chars>: Field.IsValidChar(<ch>), <ch> in <chars>
direct = '[' <chars> ']'
macro = '@' <chars>
Sort = '#' <number>

Term0 = <EQ> <value> | <value> | <direct> | <macro> | <mask> |
        <value> <~> <value> | <(> <Term1> <)>
Term1 = <!> <Term0> | <Term0> | <Term0> <|> <Term1> |
        <Term0> <&> <Term1> | <Sort> <Term1>

==================================================================
Valid propertys (DataSource.DataSet, Language, ApplyMode)
combinations:
==================================================================
DS. TYPE --     Q T      Q T     Q T
LANG / ---- SQL + +      + -     + +
     \ ---- FLT - -      + +     + +
APPLY  ----     REPL     MODIFY  NOTIFY
}

// ??? Change allways there need TDataSet -> TDBDataSet
// ??? and remove (may be) TBDEDataSet

{
  ==================================================================
  DEFINE NCQBFLOOKUP for fkLookup fields support
  ==================================================================
}

type
    TNCQBFTranslator = class;
    TQBFLink = class;
    TNCQBF = class;

    TQBFState = (qsInactive, qsEdit, qsActive);
    TQBFLanguage = (qlSQL, qlFilter);
    TQBFApplyMode = (qaReplace, qaModify, qaNotify);
    TQBFToken = (tkEQ, tkGR, tkLO, tkGE, tkLE, tkNOTEQ, tkVALUE, tkMASK,
                 tkBETWEEN, tkLPAR, tkRPAR, tkOR, tkAND, tkEND, tkDIRECT,
                 tkSORTASC, tkSORTDESC, tkNOT,
                 { next is used for QBF builder }
                 tkNULL, tkNOTNULL, tkSMASK, tkEMASK, tkIMASK);
    TQBFTokens = set of TQBFToken;
    TQBFOption = (qoNoCase, qoNoDirect, qoDirectAsValue, qoDistinct, qoSQLUseRange,
                  qoSQLUseFilter, qoSQLLive, qoSQLCloneFields, qoSQLCloneCalcs,
                  qoSQLCloneEvents, qoSQLPreview, qoSQLCachedUpdates);
    TQBFOptions = set of TQBFOption;
    TSQLLookupTemplate = (ltExists, ltIn, ltJoin);
    TQBFIfPostError = (qeCancel, qeEdit, qeError);
    TQBFBuilderInterface = (qiInplace, qiDialog, qiExternal);
    TQBFCalcOperation = (coSUM, coCOUNT, coAVERAGE, coMIN, coMAX, coDISTINCT);
    TQBFCalcOperations = set of TQBFCalcOperation;

    TDetailEvent = procedure (Sender: TObject; AActive: Boolean) of object;
    TLocateEvent = procedure (Sender: TObject; ARestart: Boolean; ADir: Integer;
        var AFound: Boolean) of object;
    TCalculateEvent = procedure (Sender: TObject; var AValue: String) of object;
    TRenderCondEvent = procedure (Sender: TObject; var AValue: String) of object;
    TEditCondEvent = procedure (Sender: TObject; var ADefEdit, AModified: Boolean) of object;
    TValueRetranEvent = function (ADisplayValue: Variant): Variant of object;

    TNCQBF = class(TComponent)
    private
        FQBFTable: TNCMemTable;
        FQBFState: TQBFState;
        FMacros: TStrings;
        FSQL: TStringList;
        FSQLOrderBy: String;
        FSQLFrom: String;
        FTmpQuery: TQuery;
        FQBFOptions: TQBFOptions;
        FQBFApplyMode: TQBFApplyMode;
        FSQLAlias: String;
        FDataSource: TDataSource;
        FDetailSource: TDataSource;
        FDetailActive: Boolean;
        FQBFIfPostError: TQBFIfPostError;
        FQBFBuilderInterface: TQBFBuilderInterface;
        FQBFLanguage: TQBFLanguage;
        FKeyFields: String;
        FFirstLink: TQBFLink;
        FShowHint: Boolean;
        FSQLServer: TNCSQLServer;
        FSQLLookupTemplate: TSQLLookupTemplate;
        FLoopEdit: Boolean;

        // Saved DataSet parameters
        FPrevDataSet: TBDEDataSet;
        FPrevActive: Boolean;
        FPrevDataSetChanged: Boolean;
        FPrevPosition: Variant;
        FPrevIndexFieldNames: String;
        FPrevIndexName: String;
        FPrevFilterOptions: TFilterOptions;
        FPrevFilter: String;
        FPrevFiltered: Boolean;
        FPrevRng: TNCRangeStore;
        FLastIndexField: Integer;
        FPrevSQL: String;

        // Locate & Calculate
        FOperationField: TField;
        FLocateMask: String;
        FCalculateOperation: TQBFCalcOperations;

        // Events
        FOnBeforeCancel: TNotifyEvent;
        FOnAfterCancel: TNotifyEvent;
        FOnBeforeEdit: TNotifyEvent;
        FOnAfterEdit: TNotifyEvent;
        FOnBeforePost: TNotifyEvent;
        FOnAfterPost: TNotifyEvent;
        FOnDetail: TDetailEvent;
        FOnLocate: TLocateEvent;
        FOnCalculate: TCalculateEvent;
        FOnRenderCond: TRenderCondEvent;
        FOnEditCond: TEditCondEvent;

        procedure SetMacros(m: TStrings);
        procedure SetDataSource(AValue: TDataSource);
        procedure SetState(st: TQBFState);
        procedure RestoreState;
        procedure SaveState;
        function DataSetValid: Boolean;
        function DataSetActive: Boolean;
        function PrevDataSetActive: Boolean;
        procedure RestoreDetailSource;
        procedure CloseDetailSource;
        function GetCanJump: Boolean;
        procedure SetOptions(AOptions: TQBFOptions);
        procedure SetLanguage(ALanguage: TQBFLanguage);
        procedure BuildDataSetSQL(ADS: TBDEDataSet; AAlias: String;
            var AFromSQL, AWhereSQL, AOrderBySQL: String;
            var AMasterDS: TDataSource; AParams: TParams);
        procedure AssembleSQL(ADestQuery: TQuery; AForDS: TBDEDataSet;
            ASelect: String; AUseOrderBy: Boolean);
        procedure AssembleFilter(AForDS: TDataSet);
        function BuildDispFldSQL(AFields, AFormat, AAlias: String;
            AMacros: TStrings): String;
        procedure BuildFieldSQL(AField: TField; var FieldPreSQL, FieldNameSQL,
            FieldPostSQL, FieldFromSQL: String; var FieldDataType: TFieldType;
            var FieldValueRetran: TValueRetranEvent);
        procedure BuildLookupSQL(AKeyFields, ALookupKeyFields, ALookupDisplayFields: String;
            ALookupMacros: TStrings; ALookupWhere: String; ALookupDS, ALookupAlias: String;
            var FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL: String);
        function GetCanLocate: Boolean;
        function GetBaseDataSet: TBDEDataSet;
        procedure StateChanged;
        procedure AddLink(ALink: TQBFLink);
        procedure RemoveLink(ALink: TQBFLink);
        procedure SetDetailSource(AValue: TDataSource);
        procedure SetSQLServerType(AValue: TSQLServerType);
        function GetSQLServerType: TSQLServerType;
        procedure SetApplyMode(AMode: TQBFApplyMode);
        procedure CheckInactive;
        procedure CheckValidProps;

    protected
        procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;

        procedure Cancel(sync: Boolean);
        procedure Edit;
        procedure Post;
        procedure Locate(reset: Boolean);
        procedure DoLocate(restart: Boolean; dir: Integer);
        procedure Calculate;
        function DoCalculate: String;
        function SQLSelect4Calc(AExpr: String): String;
        function CanQueryDS(ds: TDataSet): Boolean;
        function CanQueryField(fld: TField): TQBFTokens;

        property SQLServer: TNCSQLServer read FSQLServer write FSQLServer;
        property CanQuery: Boolean read DataSetValid;
        property CanJump: Boolean read GetCanJump;
        property CanFirstLocate: Boolean read DataSetActive;
        property CanNextLocate: Boolean read GetCanLocate;
        property CanCalculate: Boolean read DataSetActive;

        property SQLFrom: String read FSQLFrom;
        property SQLOrderBy: String read FSQLOrderBy;
        property SQL: TStringList read FSQL;
        property State: TQBFState read FQBFState;

        property QBFContainer: TNCMemTable read FQBFTable;

        property OperationField: TField read FOperationField write FOperationField;
        property LocateMask: String read FLocateMask write FLocateMask;
        property CalculateOperation: TQBFCalcOperations read FCalculateOperation
            write FCalculateOperation;

    published
        property DataSource: TDataSource read FDataSource write SetDataSource;
        property DetailSource: TDataSource read FDetailSource write SetDetailSource;
        property IfPostError: TQBFIfPostError read FQBFIfPostError write FQBFIfPostError
            default qeEdit;
        property KeyFields: String read FKeyFields write FKeyFields;
        property Language: TQBFLanguage read FQBFLanguage write SetLanguage
            default qlSQL;
        property Macros: TStrings read FMacros write SetMacros;
        property ApplyMode: TQBFApplyMode read FQBFApplyMode write SetApplyMode
            default qaReplace;
        property Options: TQBFOptions read FQBFOptions write SetOptions
            default [qoNoCase, qoNoDirect, qoSQLUseRange, qoSQLUseFilter,
                     qoSQLCloneFields, qoSQLCloneCalcs, qoSQLCloneEvents,
                     qoSQLCachedUpdates];
        property QueryBuilder: TQBFBuilderInterface read FQBFBuilderInterface
            write FQBFBuilderInterface default qiDialog;
        property SQLServerType: TSQLServerType read GetSQLServerType write SetSQLServerType
{$IFDEF NCHOME}
            default stOracle;
{$ELSE}
            default stLocal;
{$ENDIF}
        property SQLLookupTemplate: TSQLLookupTemplate read FSQLLookupTemplate
            write FSQLLookupTemplate default ltIn;
        property SQLAlias: String read FSQLAlias write FSQLAlias;
        property ShowHint: Boolean read FShowHint write FShowHint
            default True;

        property OnBeforeCancel: TNotifyEvent read FOnBeforeCancel write FOnBeforeCancel;
        property OnAfterCancel: TNotifyEvent read FOnAfterCancel write FOnAfterCancel;
        property OnBeforeEdit: TNotifyEvent read FOnBeforeEdit write FOnBeforeEdit;
        property OnAfterEdit: TNotifyEvent read FOnAfterEdit write FOnAfterEdit;
        property OnBeforePost: TNotifyEvent read FOnBeforePost write FOnBeforePost;
        property OnAfterPost: TNotifyEvent read FOnAfterPost write FOnAfterPost;
        property OnDetail: TDetailEvent read FOnDetail write FOnDetail;
        property OnLocate: TLocateEvent read FOnLocate write FOnLocate;
        property OnCalculate: TCalculateEvent read FOnCalculate write FOnCalculate;
        property OnRenderCond: TRenderCondEvent read FOnRenderCond write FOnRenderCond;
        property OnEditCond: TEditCondEvent read FOnEditCond write FOnEditCond;
    end;

    TQBFLink = class
    private
        FQBF: TNCQBF;
        FNext: TQBFLink;
        function GetState: TQBFState;
        procedure SetQBF(AQBF: TNCQBF);
    protected
        procedure StateChanged; virtual;
    public
        destructor Destroy; override;
        property QBF: TNCQBF read FQBF write SetQBF;
        property State: TQBFState read GetState;
    end;

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }

    TQBFTranOption = (toDirectAsValue, toNoCase, toNoDirect);
    TQBFTranOptions = set of TQBFTranOption;

    TNCQBFTranslator = class
    private
        FToken: TQBFToken;
        FTokenStr: String;
        FParenthesisLevel: Integer;
        FParsePos: Integer;
        FItemQBF: String;
        FItemSQL: String;
        FItemSort: Integer;
        FLang: TQBFLanguage;
        FOptions: TQBFTranOptions;
        FMacros: TStrings;
        FEnTokens: TQBFTokens;
        FItemFieldName: String;
        FItemFieldLabel: String;
        FItemFieldType: TFieldType;
        FSQLServer: TNCSQLServer;
        FValueRetranEvent: TValueRetranEvent;

        procedure ReadToken;
        procedure ReadValue;
        function ReadTokenIF(needToken: TQBFToken; MastUndo: Boolean): Boolean;
        procedure Term1;
        procedure Term0;
        procedure WriteStr(const s: String);
        procedure WriteBinOp(const op: String);
        procedure WriteTokenStr;
        procedure WriteFieldName;
        procedure TokenAllowed(t: TQBFToken);

    public
        constructor Create;

        procedure Translate(AFieldLabel, AFieldName: String; AFieldType: TFieldType;
            AEnTokens: TQBFTokens; ACond: String; ARetranEvent: TValueRetranEvent);
        property Where: String read FItemSQL;
        property OrderBy: Integer read FItemSort;

        property Language: TQBFLanguage read FLang write FLang default qlSQL;
        property Options: TQBFTranOptions read FOptions write FOptions default [toNoCase];
        property Macros: TStrings read FMacros write FMacros;
        property SQLServer: TNCSQLServer read FSQLServer write FSQLServer;
    end;

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }

implementation

Uses NCUtil, NCUIUtil, DsgnIntf, NCQBFEdt, Forms, NCQBFExp, NCStrs, NcWait,
     NCSelFld, NCQBFLoc, NCQBFClc, Consts, BDE, NCRefFld, NCBrDB, NCBrArr,
     Controls;

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }

constructor TNCQBF.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    FSQL := TStringList.Create;
    FMacros := TStringList.Create;
    FQBFOptions := [qoNoCase, qoNoDirect, qoSQLUseRange, qoSQLUseFilter,
                    qoSQLCloneFields, qoSQLCloneCalcs, qoSQLCloneEvents,
                    qoSQLCachedUpdates];
    FQBFApplyMode := qaReplace;
    FQBFBuilderInterface := qiDialog;
    FShowHint := True;
    FSQLServer := TNCSQLServer.Create;
    FSQLServer.ServerType :=
{$IFDEF NCHOME}
            stOracle;
{$ELSE}
            stLocal;
{$ENDIF}
    FSQLLookupTemplate := ltIn;
    FQBFIfPostError := qeEdit;
    FTmpQuery := TQuery.Create(nil);
    FSQLAlias := 'NCQ';
end;

destructor TNCQBF.Destroy;
begin
    Cancel(False);
    while FFirstLink <> nil do
        RemoveLink(FFirstLink);
    FMacros.Free;
    FSQL.Free;
    FSQLServer.Free;
    FTmpQuery.Free;
    FTmpQuery := nil;
    inherited Destroy;
end;

procedure TNCQBF.Notification(AComponent: TComponent; Operation: TOperation);
begin
    inherited Notification(AComponent, Operation);
    if (Operation = opRemove) then begin
        if (AComponent = DataSource) then
            DataSource := nil;
        if (AComponent = DetailSource) then
            DetailSource := nil;
        if (AComponent = FPrevDataSet) then
            FPrevDataSet := nil;
        if (AComponent = FOperationField) then
            FOperationField := nil;
    end;
end;

procedure TNCQBF.SetSQLServerType(AValue: TSQLServerType);
begin
    CheckInactive;
    FSQLServer.ServerType := AValue;
end;

function TNCQBF.GetSQLServerType: TSQLServerType;
begin
    Result := FSQLServer.ServerType;
end;

procedure TNCQBF.SetDataSource(AValue: TDataSource);
begin
    CheckInactive;
    if FDataSource <> AValue then begin
        if (NCQBFLocateDlg <> nil) and NCQBFLocateDlg.Visible and
           (NCQBFLocateDlg.QBF = Self) then
            NCQBFLocateDlg.Hide;
        FDataSource := AValue;
        if AValue <> nil then
            AValue.FreeNotification(Self);
        if (KeyFields = '') and DataSetActive and (AValue.DataSet is TTable) then
            KeyFields := DUUniqueFields(TTable(AValue.DataSet));
        StateChanged;
    end;
end;

procedure TNCQBF.SetDetailSource(AValue: TDataSource);
begin
    CheckInactive;
    FDetailSource := AValue;
    if AValue <> nil then
        AValue.FreeNotification(Self);
end;

procedure TNCQBF.SetMacros(m: TStrings);
begin
    CheckInactive;
    FMacros.Assign(m);
end;

procedure TNCQBF.SetApplyMode(AMode: TQBFApplyMode);
begin
    CheckInactive;
    if ApplyMode <> AMode then begin
        if (ApplyMode = qaReplace) then begin
            FTmpQuery.Free;
            FTmpQuery := nil;
        end;
        FQBFApplyMode := AMode;
        if AMode = qaReplace then
            Language := qlSQL;
        if (ApplyMode = qaReplace) then
            FTmpQuery := TQuery.Create(nil);
    end;
end;

procedure TNCQBF.SetOptions(AOptions: TQBFOptions);
begin
    CheckInactive;
    FQBFOptions := AOptions;
end;

procedure TNCQBF.SetLanguage(ALanguage: TQBFLanguage);
begin
    CheckInactive;
    FQBFLanguage := ALanguage;
end;

procedure TNCQBF.SetState(st: TQBFState);
begin
    FQBFState := st;
    StateChanged;
end;

function TNCQBF.GetBaseDataSet: TBDEDataSet;
begin
    if State = qsInactive then
        Result := DataSource.DataSet as TBDEDataSet
    else
        Result := FPrevDataSet;
end;

function TNCQBF.DataSetValid: Boolean;
begin
    Result := (DataSource <> nil) and (DataSource.DataSet <> nil) and
        not (csDestroying in DataSource.ComponentState) and
        not (csDestroying in DataSource.DataSet.ComponentState);
end;

function TNCQBF.DataSetActive: Boolean;
begin
    Result := DataSetValid and DataSource.DataSet.Active;
end;

function TNCQBF.PrevDataSetActive: Boolean;
begin
    Result := (FPrevDataSet <> nil) and FPrevDataSet.Active and
        not (csDestroying in FPrevDataSet.ComponentState);
end;

function TNCQBF.GetCanJump: Boolean;
begin
    Result := PrevDataSetActive and DataSetActive and (State = qsActive) and
        (KeyFields <> '');
end;

procedure TNCQBF.CheckInactive;
begin
    if State <> qsInactive then
        GenErrorRes(SMastBeInactive);
end;

procedure TNCQBF.CheckValidProps;
begin
    if not DataSetValid then
        GenErrorRes(SNeedDataSource);
    if (ApplyMode = qaReplace) and (Language <> qlSQL) or
       (ApplyMode = qaModify) and (Language = qlSQL) and
       not (DataSource.DataSet is TQuery) then
        GenErrorRes(SInvalidPropertySet);
    with DataSource.DataSet do
        if Active then
            CheckBrowseMode;
end;

procedure TNCQBF.SaveState;

    procedure LocateLastField;
    var
        i: Integer;
    begin
        with TTable(FPrevDataSet) do
        for i := IndexFieldCount - 1 downto 0 do
            if (FLastIndexField < i) and not IndexFields[i].IsNull then begin
                FLastIndexField := i;
                Exit;
            end;
    end;

begin
    FPrevDataSetChanged := False;
    if (ApplyMode <> qaModify) then
        Exit;
    if (FPrevDataSet is TTable) and (Language = qlFilter) then
        with TTable(FPrevDataSet) do begin
            FPrevIndexFieldNames := IndexFieldNames;
            FPrevIndexName := IndexName;
            if MasterSource = nil then
                DUSaveRestRanges(TTable(FPrevDataSet), FPrevRng, True)
            else begin
                FPrevRng.FStRange := Unassigned;
                FPrevRng.FEnRange := Unassigned;
            end;
            FLastIndexField := -1;
            if MasterSource <> nil then
                FLastIndexField := DUCountField(MasterFields) - 1
            else if IndexFieldCount <> 0 then
                try
                    EditRangeStart;
                    LocateLastField;
                    EditRangeEnd;
                    LocateLastField;
                finally
                    Cancel;
                end;
        end
    else if (FPrevDataSet is TQuery) and (Language = qlSQL) then
        FPrevSQL := TQuery(FPrevDataSet).SQL.Text;
    if Language = qlFilter then begin
        FPrevFilterOptions := FPrevDataSet.FilterOptions;
        FPrevFilter := FPrevDataSet.Filter;
        FPrevFiltered := FPrevDataSet.Filtered;
    end;
    FPrevActive := FPrevDataSet.Active;
end;

procedure TNCQBF.RestoreState;
begin
    FSQL.Clear;
    FSQLOrderBy := '';
    FSQLFrom := '';
    if (ApplyMode = qaReplace) and (FTmpQuery <> nil) then begin
        FTmpQuery.Close;
        FTmpQuery.SQL.Clear;
    end
    else if (ApplyMode = qaModify) and (FPrevDataSet <> nil) and FPrevDataSetChanged then begin
        FPrevDataSet.Close;
        if (Language = qlFilter) then begin
            FPrevDataSet.FilterOptions := FPrevFilterOptions;
            FPrevDataSet.Filter := FPrevFilter;
            FPrevDataSet.Filtered := FPrevFiltered;
        end;
        if (FPrevDataSet is TTable) and (Language = qlFilter) then
            with TTable(FPrevDataSet) do begin
                if FPrevIndexFieldNames <> '' then
                    IndexFieldNames := FPrevIndexFieldNames
                else
                    IndexName := FPrevIndexName;
            end
        else if (FPrevDataSet is TQuery) and (Language = qlSQL) then
            TQuery(FPrevDataSet).SQL.Text := FPrevSQL;
        if FPrevActive then begin
            FPrevDataSet.Open;
            try
                if (FPrevDataSet is TTable) and (TTable(FPrevDataSet).MasterSource = nil) then
                    DUSaveRestRanges(TTable(FPrevDataSet), FPrevRng, False);
                if not VarIsEmpty(FPrevPosition) then begin
                    FPrevDataSet.Locate(KeyFields, FPrevPosition, []);
                    FPrevPosition := Unassigned;
                end;
            except
                Application.HandleException(Self);
            end;
        end;
    end;
    FPrevDataSetChanged := False;
end;

procedure TNCQBF.CloseDetailSource;
begin
    if (DetailSource <> nil) and (DetailSource.DataSet <> nil) then begin
        FDetailActive := DetailSource.DataSet.Active;
        DetailSource.DataSet.Active := False;
    end;
    if Assigned(FOnDetail) then
        FOnDetail(Self, False);
end;

procedure TNCQBF.RestoreDetailSource;
begin
    if (DetailSource <> nil) and (DetailSource.DataSet <> nil) then
        DetailSource.DataSet.Active := FDetailActive;
    if Assigned(FOnDetail) then
        FOnDetail(Self, True);
end;

function TNCQBF.CanQueryDS(ds: TDataSet): Boolean;
begin
    Result := (ds is TDBDataSet) and (TDBDataSet(ds).DatabaseName <> '') and
        ((ds is TTable) or
         (ds is TQuery) and SQLServer.Can(scFromQuery));
end;

function TNCQBF.CanQueryField(fld: TField): TQBFTokens;
const
    blobFld: TQBFTokens = [tkEQ, tkNOTEQ, tkLPAR..tkDIRECT, tkNULL..tkNOTNULL, tkNOT];
    sqlFld: TQBFTokens = [tkEQ..tkDIRECT, tkNULL..tkIMASK, tkNOT];
    fltFld: TQBFTokens = [tkEQ..tkVALUE, tkBETWEEN..tkDIRECT, tkNULL..tkNOTNULL,
        tkSMASK, tkNOT];
    fltRefFld: TQBFTokens = [tkEQ, tkNOTEQ, tkVALUE, tkLPAR, tkRPAR, tkOR, tkAND,
        tkEND, tkNULL, tkNOTNULL];

    function CanQueryReferField(fld: TNCReferField): Boolean;
    begin
        Result := not (roNoQBF in fld.RefOptions);
        if not Result then
            Exit;
        if (Language = qlSQL) then begin
            if (fld.ReferKind = rkBrowser) and (fld.RefBrowser is TNCArrBrowser) then
                with TNCArrBrowser(fld.RefBrowser) do
                    Result := (LookupKind in [lkDataSet {$IFDEF NCQBFLOOKUP}, lkLookupField {$ENDIF}]) and
                                    CanQueryDS(SourceDataSet) or
                              not (LookupKind in [lkDataSet {$IFDEF NCQBFLOOKUP}, lkLookupField {$ENDIF}]) and
                                    SQLServer.Can(scFromValues)
            else if fld.ReferKind = rkPickIndexedList then
                Result := SQLServer.Can(scDecode);
        end
        else if (Language = qlFilter) then begin
            Result := (TNCReferField(fld).ReferKind in [rkBrowser, rkPickList, rkPickIndexedList]) and
                      (DUCountField(TNCReferField(fld).KeyFields) = 1);
        end;
    end;

var
    i: Integer;
begin
    if not fld.Visible or
       not RFCheckRefer(fld) and (fld.FieldKind = fkCalculated) and
            (FMacros.Values[AnsiUpperCase(fld.FieldName)] = '') or
       RFCheckRefer(fld) and
            not ((Language = qlSQL) and not (qoSQLLive in Options) and
                    ((SQLAlias <> '') or (SQLLookupTemplate = ltIn)) and
                    CanQueryReferField(TNCReferField(fld)) or
                 (Language = qlFilter) and (TNCReferField(fld).ReferKind in
                    [rkBrowser, rkPickList, rkPickIndexedList]) and
                    CanQueryReferField(TNCReferField(fld))
                )
{$IFDEF NCQBFLOOKUP}
       or (fld.FieldKind = fkLookup) and
           ((Language <> qlSQL) or (qoSQLLive in Options) or
            (SQLAlias = '') and (SQLLookupTemplate <> ltIn) or
            not CanQueryDS(fld.LookupDataSet))
{$ENDIF}
       then
        Result := []
    else if fld.DataType in DUBlobFieldTypes then
        Result := blobFld
    else if Language = qlSQL then begin
        Result := sqlFld;
        if (fld.FieldKind = fkData) or
           not (qoSQLLive in Options) and
           (RFCheckRefer(fld) {$IFDEF NCQBFLOOKUP} or (fld.FieldKind = fkLookup) {$ENDIF}) and
           (SQLLookupTemplate = ltJoin) and SQLServer.Can(scOrderByExp) then
            Result := Result + [tkSORTASC, tkSORTDESC];
    end
    else if Language = qlFilter then begin
        if RFCheckRefer(fld) then
            Result := fltRefFld
        else begin
            Result := fltFld;
            if (FPrevDataSet is TTable) and (fld.FieldKind = fkData) then
                with TTable(FPrevDataSet) do begin
                    Include(Result, tkSORTASC);
                    for i := 0 to FLastIndexField do
                        if IndexFields[i] = fld then begin
                            Exclude(Result, tkSORTASC);
                            Break;
                        end;
                end;
        end;
    end;
end;

procedure TNCQBF.StateChanged;
var
    ALink: TQBFLink;
begin
    ALink := FFirstLink;
    while ALink <> nil do begin
        ALink.StateChanged;
        ALink := ALink.FNext;
    end;
end;

procedure TNCQBF.AddLink(ALink: TQBFLink);
begin
    ALink.FQBF := Self;
    ALink.FNext := FFirstLink;
    FFirstLink := ALink;
    ALink.StateChanged;
end;

procedure TNCQBF.RemoveLink(ALink: TQBFLink);
var
    prev, cur: TQBFLink;
begin
    prev := nil;
    cur := FFirstLink;
    while (cur <> nil) and (cur <> ALink) do begin
        prev := cur;
        cur := cur.FNext;
    end;
    if cur = FFirstLink then
        FFirstLink := cur.FNext
    else
        prev.FNext := cur.FNext;
    ALink.FQBF := nil;
    if not (csDestroying in ComponentState) then
        ALink.StateChanged;
end;

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }

procedure TNCQBF.Cancel(sync: Boolean);
begin
    if State = qsInactive then
        Exit;
    if not (csDestroying in ComponentState) and Assigned(FOnBeforeCancel) then
        FOnBeforeCancel(Self);
    WDSuspend(False, '');
    try
        if DataSetActive then
            DataSource.DataSet.Cancel;
        if sync and CanJump then
        try
            case ApplyMode of
            qaReplace, qaNotify:
                FPrevDataSet.Locate(KeyFields, DataSource.DataSet[KeyFields], []);
            qaModify:
                FPrevPosition := DataSource.DataSet[KeyFields];
            end;
        except
            Application.HandleException(Self);
        end;
        try
            RestoreState;
        finally
            if not (csDestroying in DataSource.ComponentState) and
               (DataSource.DataSet <> FPrevDataSet) then
                DataSource.DataSet := FPrevDataSet;
            if (QueryBuilder = qiInplace) and (State = qsEdit) then
                RestoreDetailSource;
        end;
    finally
        FQBFTable.Free;
        SetState(qsInactive);
        WDResume;
    end;
    if not (csDestroying in ComponentState) and Assigned(FOnAfterCancel) then
        FOnAfterCancel(Self);
end;

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }

procedure TNCQBF.Edit;
    { ----------------------------------------- }
    procedure CreateQBFTable;
    var
        i: Integer;
        fldDS, fldQBF: TField;
    begin
        FQBFTable := TNCMemTable.Create(nil);
        with FPrevDataSet do
             for i := 0 to FieldCount - 1 do
                  FQBFTable.FieldDefs.Add(Fields[i].FieldName, ftString, 100, False);
        with FQBFTable do begin
            MarkDelete := True;
            CreateTable;
            Open;
            for i := 0 to FieldCount - 2 do begin
                fldDS := FPrevDataSet.Fields[i];
                fldQBF := FieldByName(fldDS.FieldName);
                fldQBF.DisplayWidth := fldDS.DisplayWidth;
                fldQBF.DisplayLabel := fldDS.DisplayLabel;
                fldQBF.Visible := fldDS.Visible;
                fldQBF.Index := fldDS.Index;
                fldQBF.ReadOnly := (CanQueryField(fldDS) = []);
            end;
            FieldByName(MTDelMark).Visible := False;
        end;
    end;

    procedure RunDialogIterface;
    begin
        if NCQBFEditDlg = nil then
            NCQBFEditDlg := TNCQBFEditDlg.Create(Application);
        if NCQBFEditDlg.Execute(FPrevDataSet, FQBFTable, Self) then
            Post
        else
            Cancel(False);
    end;

begin
    if State = qsEdit then
        Exit;
    if State = qsInactive then
        CheckValidProps;
    repeat
        FLoopEdit := False;
        try
            if Assigned(FOnBeforeEdit) then
                FOnBeforeEdit(Self);
            if State = qsInactive then begin
                FPrevDataSet := DataSource.DataSet as TBDEDataSet;
                SaveState;
                CreateQBFTable;
            end
            else if State = qsActive then begin
                RestoreState;
                FQBFTable.First;
            end;
            if QueryBuilder = qiInplace then
                CloseDetailSource;
            FQBFTable.Edit;
            if QueryBuilder = qiInplace then
                DataSource.DataSet := FQBFTable;
            SetState(qsEdit);
            if Assigned(FOnAfterEdit) then
                FOnAfterEdit(Self);
        except
            Cancel(False);
            raise;
        end;
        if QueryBuilder = qiDialog then
            RunDialogIterface;
    until not FLoopEdit;
end;

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }
{ DataSet Query Building Engine }

procedure TNCQBF.BuildDataSetSQL(ADS: TBDEDataSet; AAlias: String;
    var AFromSQL, AWhereSQL, AOrderBySQL: String;
    var AMasterDS: TDataSource; AParams: TParams);
begin
    AFromSQL := DUObjName2SQL(ADS, AAlias);
    AWhereSQL := '';
    AOrderBySQL := '';
    AMasterDS := nil;
    if AParams <> nil then
        AParams.Clear;
    if ADS is TTable then
        with TTable(ADS) do begin
            if qoSQLUseRange in FQBFOptions then begin
                if MasterSource <> nil then begin
                    AWhereSQL := DUJoin2SQL(DUIndex2Fields(TTable(ADS)), MasterFields,
                        AAlias, ':', nil, nil);
                    AMasterDS := MasterSource;
                end
                else if ADS.Active then
                    AWhereSQL := DUTblRange2SQL(TTable(ADS), AAlias);
            end;
            if IndexFieldCount > 0 then
                AOrderBySQL := DUFields2SQL(DUIndex2Fields(TTable(ADS)),
                    AAlias, False, Macros);
        end
    else if ADS is TQuery then begin
        if (qoSQLUseRange in FQBFOptions) and (TQuery(ADS).DataSource <> nil) then
            AMasterDS := TQuery(ADS).DataSource;
        if AParams <> nil then
            AParams.Assign(TQuery(ADS).Params);
    end;
    if (qoSQLUseFilter in FQBFOptions) and (ADS.Filter <> '') and ADS.Filtered then begin
        if AWhereSQL <> '' then
            AWhereSQL := AWhereSQL + ' and ';
        AWhereSQL := AWhereSQL + '(' + DUFilter2SQL(ADS.Filter, AAlias) + ')';
    end;
end;

procedure TNCQBF.AssembleSQL(ADestQuery: TQuery; AForDS: TBDEDataSet;
    ASelect: String; AUseOrderBy: Boolean);
var
    wasWhere: Boolean;
    s, dsSelect, dsFrom, dsWhere, dsOrderBy: String;
    pars: TParams;
    dsMasterDS: TDataSource;
begin
    pars := nil;
    if ADestQuery <> AForDS then
        pars := ADestQuery.Params;
    BuildDataSetSQL(AForDS, SQLAlias, dsFrom, dsWhere, dsOrderBy, dsMasterDS, pars);
    with ADestQuery do
    try
        SQL.BeginUpdate;
        SQL.Clear;
        dsSelect := ASelect;
        if qoDistinct in Options then
            dsSelect := 'distinct ' + dsSelect;
        SQL.Add('select ' + dsSelect);
        if FSQLFrom <> '' then
            dsFrom := dsFrom + ', ' + FSQLFrom;
        SQL.Add('from ' + dsFrom);
        wasWhere := False;
        if dsWhere <> '' then
            DUContWhere(SQL, wasWhere, dsWhere);
        if FSQL.Count <> 0 then
            if Language = qlSQL then
                DUContWhere(SQL, wasWhere, FSQL.Text)
            else { if (ApplyMode = qaNotify) then }
                DUContWhere(SQL, wasWhere, DUFilter2SQL(FSQL.Text, SQLAlias));
        if AUseOrderBy then begin
            s := 'order by ';
            if (SQLOrderBy = '') then begin
                if dsOrderBy <> '' then
                    SQL.Add(s + dsOrderBy);
            end
            else if Language = qlSQL then
                SQL.Add(s + SQLOrderBy)
            else
                SQL.Add(s + DUFields2SQL(SQLOrderBy, SQLAlias, False, nil));
        end;
        DataSource := dsMasterDS;
        SQL.Text := SQL.Text; 
    finally
        SQL.EndUpdate;
    end;
end;

procedure TNCQBF.AssembleFilter(AForDS: TDataSet);
var
    i: Integer;
    s: String;
begin
    if (AForDS is TTable) and (FSQLOrderBy <> '') then
        with TTable(AForDS) do begin
            s := '';
            for i := 0 to FLastIndexField do
                s := s + IndexFields[i].FieldName + ';';
        end;
    AForDS.Close;
    if FSQL.Count <> 0 then begin
        if AForDS.Filter <> '' then
            AForDS.Filter := '(' + AForDS.Filter + ') and ' + FSQL.Text
        else
            AForDS.Filter := FSQL.Text;
//        AForDS.FilterOptions := [foNoPartialCompare];
        if qoNoCase in Options then
            AForDS.FilterOptions := AForDS.FilterOptions + [foCaseInsensitive];
        AForDS.Filtered := True;
    end;
    if (AForDS is TTable) then
        with TTable(AForDS) do begin
            if FSQLOrderBy <> '' then
                IndexFieldNames := s + FSQLOrderBy;
            Open;
            if MasterSource = nil then
                DUSaveRestRanges(TTable(AForDS), FPrevRng, False);
        end
    else
        AForDS.Open;
end;

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }
{ Field Query Building Engine }

procedure TNCQBF.BuildLookupSQL(AKeyFields, ALookupKeyFields, ALookupDisplayFields: String;
    ALookupMacros: TStrings; ALookupWhere: String; ALookupDS, ALookupAlias: String;
    var FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL: String);

    procedure AppendLookupWhere;
    begin
        if ALookupWhere <> '' then
            FieldPreSQL := FieldPreSQL + ALookupWhere + ') and (';
    end;

begin
    case SQLLookupTemplate of
    ltExists:
        begin
            FieldPreSQL := 'exists (select Null from ' + ALookupDS + ' where (';
            AppendLookupWhere;
            FieldPreSQL := FieldPreSQL + DUJoin2SQL(AKeyFields, ALookupKeyFields,
                SQLAlias, ALookupAlias, nil, ALookupMacros) + ') and (';
            FieldPostSQL := '))';
        end;
    ltIn:
        begin
            FieldPreSQL := '(' + DUFields2SQL(AKeyFields, SQLAlias, False, nil) +
                ') in (select ' + DUFields2SQL(ALookupKeyFields, ALookupAlias, False, ALookupMacros) +
                ' from ' + ALookupDS + ' where (';
            AppendLookupWhere;
            FieldPostSQL := '))';
        end;
    ltJoin:
        if SQLServer.Can(scFromQuery) then begin
            FieldFromSQL := '(select ' + DUFields2SQL(DUJoinFields(ALookupDisplayFields,
                ALookupKeyFields), '', True, ALookupMacros) + ' from ' + ALookupDS;
            if ALookupWhere <> '' then
                FieldFromSQL := FieldFromSQL + ' where ' + ALookupWhere;
            FieldFromSQL := FieldFromSQL + ') ' + ALookupAlias;
            FieldPreSQL := DUJoin2SQL(AKeyFields, ALookupKeyFields,
                SQLAlias, ALookupAlias, nil, nil) + ' and (';
            FieldPostSQL := ')';
        end
        else begin
            FieldPreSQL := '(';
            AppendLookupWhere;
            FieldPreSQL := FieldPreSQL + DUJoin2SQL(AKeyFields, ALookupKeyFields,
                SQLAlias, ALookupAlias, nil, ALookupMacros) + ') and (';
            FieldPostSQL := ')';
            FieldFromSQL := ALookupDS;
        end;
    end;
end;

function TNCQBF.BuildDispFldSQL(AFields, AFormat, AAlias: String; AMacros: TStrings): String;
var
    i: Integer;
    s: String;
begin
    if AFormat = '' then
        Result := DUFields2SQL(DUFieldName(AFields, 1), AAlias, False, AMacros)
    else begin
        Result := '';
        i := 1;
        while i <= Length(AFormat) do begin
            if (AFormat[i] = '%') and (i < Length(AFormat)) and (AFormat[i+1] <> '%') then begin
                Inc(i);
                s := DUFieldName(AFields, Ord(AFormat[i]) - Ord('0') + 1);
                s := DUFields2SQL(s, AAlias, False, AMacros);
                if s = '' then
                    GenErrorResFmt(SDisplayFormatError, [AFormat]);
                if not SQLServer.Can(scAutoCast) then
                    s := SQLServer.CastToStr(s);
                Result := Result + ''' || ' + s + ' || ''';
            end
            else
                Result := Result + AFormat[i];
            Inc(i);
        end;
        Result := '''' + Result + '''';
    end;
end;

procedure TNCQBF.BuildFieldSQL(AField: TField; var FieldPreSQL, FieldNameSQL,
    FieldPostSQL, FieldFromSQL: String; var FieldDataType: TFieldType;
    var FieldValueRetran: TValueRetranEvent);
var
    ldsFrom, ldsWhere, ldsOrderBy, tmpAlias, s: String;
    i, i1: Integer;
    ldsMaster: TDataSource;

    procedure CheckSingleField(AFields: String);
    begin
        if Pos(';', AFields) <> 0 then
            GenErrorResFmt(SMastBeSingleField, [AFields]);
    end;

begin
    FieldPreSQL := '';
    FieldNameSQL := '';
    FieldPostSQL := '';
    FieldFromSQL := '';
    s := Macros.Values[AnsiUpperCase(AField.FieldName)];
    tmpAlias := 'ncq' + IntToStr(Random(1000));
    FieldDataType := AField.DataType;
    FieldValueRetran := nil;
    if s = '' then begin
        case AField.FieldKind of
{$IFDEF NCQBFLOOKUP}
        fkLookup:
            begin
                CheckSingleField(AField.LookupResultField);
                FieldNameSQL := tmpAlias + '.' + AField.LookupResultField;
                if (SQLLookupTemplate = ltJoin) and SQLServer.Can(scFromQuery) then
                    BuildDataSetSQL(AField.LookupDataSet as TBDEDataSet,
                        '', ldsFrom, ldsWhere, ldsOrderBy, ldsMaster, nil)
                else
                    BuildDataSetSQL(AField.LookupDataSet as TBDEDataSet,
                        tmpAlias, ldsFrom, ldsWhere, ldsOrderBy, ldsMaster, nil);
                if ldsMaster = nil then
                    BuildLookupSQL(AField.KeyFields, AField.LookupKeyFields,
                        AField.LookupResultField, nil, ldsWhere, ldsFrom, tmpAlias,
                        FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL);
            end;
{$ENDIF}
        fkData:
            begin
                FieldNameSQL := AField.FieldName;
                if (SQLAlias <> '') and (Language <> qlFilter) then
                    FieldNameSQL := SQLAlias + '.' + FieldNameSQL;
            end;
        fkCalculated:
            if AField is TNCReferField then with TNCReferField(AField) do
            if Language = qlSQL then
                case ReferKind of
                rkBrowser:
                    if RefBrowser is TNCDBBrowser then
                        with TNCDBBrowser(RefBrowser) do
                            if (SQLLookupTemplate = ltJoin) and SQLServer.Can(scFromQuery) then begin
                                FieldNameSQL := BuildDispFldSQL(DisplayFields, RefDisplayFormat, tmpAlias, nil);
                                BuildLookupSQL(AField.KeyFields, TNCDBBrowser(RefBrowser).KeyFields,
                                    TNCDBBrowser(RefBrowser).DisplayFields, Macros, SQLWhere, SQLFrom,
                                    tmpAlias, FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL);
                            end
                            else begin
                                FieldNameSQL := BuildDispFldSQL(DisplayFields, RefDisplayFormat, '', Macros);
                                BuildLookupSQL(AField.KeyFields, TNCDBBrowser(RefBrowser).KeyFields,
                                    TNCDBBrowser(RefBrowser).DisplayFields, Macros, SQLWhere, SQLFrom,
                                    '', FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL);
                            end
                    else if RefBrowser is TNCArrBrowser then
                        with TNCArrBrowser(RefBrowser) do begin
                            FieldNameSQL := BuildDispFldSQL(DisplayFields, RefDisplayFormat, tmpAlias, nil);
                            if LookupKind in [lkDataSet {$IFDEF NCQBFLOOKUP}, lkLookupField {$ENDIF}] then begin
                                BuildDataSetSQL(SourceDataSet {$IFDEF D3} as TBDEDataSet {$ENDIF},
                                    tmpAlias, ldsFrom, ldsWhere, ldsOrderBy, ldsMaster, nil);
                                if ldsMaster = nil then
                                    BuildLookupSQL(AField.KeyFields, TNCArrBrowser(RefBrowser).KeyFields,
                                        TNCArrBrowser(RefBrowser).DisplayFields, nil, ldsWhere, ldsFrom,
                                        tmpAlias, FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL);
                            end
                            else if SQLServer.Can(scFromValues) then
                                BuildLookupSQL(AField.KeyFields, TNCArrBrowser(RefBrowser).KeyFields,
                                    TNCArrBrowser(RefBrowser).DisplayFields, nil, '',
                                    '(' + SQLServer.FromValues(LookupValues) + ') ' + tmpAlias,
                                    tmpAlias, FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL)
                            else
                                FieldNameSQL := '';
                        end;
                rkPickList:
                    begin
                        FieldNameSQL := AField.KeyFields;
                        CheckSingleField(FieldNameSQL);
                        if (SQLAlias <> '') and (Language <> qlFilter) then
                            FieldNameSQL := SQLAlias + '.' + FieldNameSQL;
                    end;
                rkPickIndexedList:
                    if SQLServer.Can(scDecode) then begin
                        FieldNameSQL := AField.KeyFields;
                        CheckSingleField(FieldNameSQL);
                        if (SQLAlias <> '') and (Language <> qlFilter) then
                            FieldNameSQL := SQLAlias + '.' + FieldNameSQL;
                        FieldNameSQL := BuildDispFldSQL(SQLServer.Decode(FieldNameSQL,
                            RefPickList) + ';' + FieldNameSQL, RefDisplayFormat, '', nil);
                    end;
                end
            else if Language = qlFilter then
                case ReferKind of
                rkBrowser, rkPickList, rkPickIndexedList:
                    begin
                        FieldNameSQL := AField.KeyFields;
                        CheckSingleField(FieldNameSQL);
                        FieldDataType := AField.DataSet.FieldByName(FieldNameSQL).DataType;
                        FieldValueRetran := TNCReferField(AField).Display2KeyValue;
                    end;
                end
        end;
        if FieldNameSQL = '' then
            GenErrorResFmt(SUnsupportedFld, [AField.FieldName]);
    end
    else begin
        { additional tables }
        i := Pos('->', s);
        if i <> 0 then begin
            FieldFromSQL := Copy(s, 1, i - 1);
            if (FieldFromSQL <> '') and (Language = qlFilter) then
                GenErrorRes(SNoMacTbl4Filter);
            s := Copy(s, i + 2, Length(s));
        end;
        { condition substitute place }
        i := Pos('{', s);
        if i = 0 then
            FieldNameSQL := s
        else begin
            i1 := Pos('}', s);
            if i1 = 0 then
                GenErrorResFmt(SExpectedEOM, [AField.FieldName]);
            FieldPreSQL := Copy(s, 1, i - 1);
            FieldPostSQL := Copy(s, i1 + 1, Length(s));
            FieldNameSQL := Copy(s, i + 1, i1 - i - 1);
        end;
    end;
end;

{
procedure TNCQBF.ExecuteFilterRecord(DataSet: TDataSet; var Accept: Boolean);
begin
end;
}

{ ------------------------------------------------------------------------- }
{ ------------------------------------------------------------------------- }

procedure TNCQBF.Post;
    { -------------------------------------- }
    procedure QBF2Query;
    const
        OrderByDelim: array[Boolean] of Char = (';', ',');
    var
        k, j, i, rNo: Integer;
        firstInRecord: Boolean;
        TmpFrom, TmpSort: TStringList;
        Tran: TNCQBFTranslator;
        FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL, s: String;
        FieldDataType: TFieldType;
        FieldValueRetran: TValueRetranEvent;
        fld: TField;
    begin
        TmpSort := TStringList.Create;
        TmpSort.Sorted := False;
        TmpFrom := TStringList.Create;
        TmpSort.Sorted := False;
        Tran := TNCQBFTranslator.Create;
        Tran.Language := Language;
        Tran.SQLServer := SQLServer;
        Tran.Options := [];
        if qoNoDirect in FQBFOptions then
            Tran.Options := Tran.Options + [toNoDirect];
        if qoDirectAsValue in FQBFOptions then
            Tran.Options := Tran.Options + [toDirectAsValue];
        if qoNoCase in FQBFOptions then
            Tran.Options := Tran.Options + [toNoCase];
        Tran.Macros := Macros;
        rNo := FQBFTable.RecNo;
        with FQBFTable do
        try
            DisableControls;
            First;
            while True do begin
                while not (RecExist or EOF) do
                    Next;
                if EOF then
                    Break;
                firstInRecord := True;
                for i := 0 to FieldCount - 1 do
                    if (AnsiCompareText(Fields[i].FieldName, MTDelMark) <> 0) and
                       not Empty(Fields[i].AsString) and not Fields[i].ReadOnly then begin
                        try
                            fld := FPrevDataSet.FieldByName(Fields[i].FieldName);
                            BuildFieldSQL(fld, FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL,
                                FieldDataType, FieldValueRetran);
                            Tran.Translate(Fields[i].DisplayLabel, FieldNameSQL, FieldDataType,
                                CanQueryField(fld), Fields[i].AsString, FieldValueRetran);
                        except
                            Fields[i].FocusControl;
                            raise;
                        end;
                        if not Empty(Tran.Where) or (Tran.OrderBy <> 0) and (FieldFromSQL <> '') then begin
                            s := Tran.Where;
                            if s = '' then
                                s := '1 = 1';
                            s := FieldPreSQL + s + FieldPostSQL;
                            if FSQL.Count > 0 then begin
                                if firstInRecord then begin
                                    firstInRecord := False;
                                    s := ') or ((' + s;
                                end
                                else
                                    s := 'and (' + s;
                            end
                            else begin
                                firstInRecord := False;
                                s := '(((' + s;
                            end;
                            FSQL.Add(s + ')');
                        end;
                        if Tran.OrderBy <> 0 then begin
                            k := Abs(Tran.OrderBy);
                            if TmpSort.Count < k then begin
                                j := k - TmpSort.Count;
                                while j > 0 do begin
                                    TmpSort.Add('');
                                    Dec(j);
                                end;
                            end;
                            TmpSort[k - 1] := FieldNameSQL;
                            TmpSort.Objects[k - 1] := TObject(Pointer(Tran.OrderBy));
                        end;
                        if not Empty(FieldFromSQL) then begin
                            k := 1;
                            while k <= Length(FieldFromSQL) do begin
                                s := ExtractFieldName(FieldFromSQL, k);
                                if TmpFrom.IndexOf(s) = -1 then
                                    TmpFrom.Add(s);
                            end;
                        end;
                    end;
                Next;
            end;
            if FSQL.Count > 0 then
                FSQL.Add('))');
            if TmpFrom.Count > 0 then
                for i := 0 to TmpFrom.Count - 1 do begin
                    if FSQLFrom <> '' then
                        FSQLFrom := FSQLFrom + ', ';
                    FSQLFrom := FSQLFrom + TmpFrom[i];
                end;
            if TmpSort.Count > 0 then
                for i := 0 to TmpSort.Count - 1 do
                    if TmpSort[i] <> '' then begin
                        if FSQLOrderBy <> '' then
                            FSQLOrderBy := FSQLOrderBy + OrderByDelim[Language = qlSQL];
                        FSQLOrderBy := FSQLOrderBy + TmpSort[i];
                        if (Language = qlSQL) and (Integer(Pointer(TmpSort.Objects[i])) < 0) then
                            FSQLOrderBy := FSQLOrderBy + ' DESC';
                    end;
            if (FSQL.Count = 0) and (TmpSort.Count = 0) then
                GenErrorRes(SEmptyQuery);
        finally
            if rNo <> -1 then
                RecNo := rNo;
            EnableControls;
            TmpSort.Free;
            TmpFrom.Free;
            Tran.Free;
        end;
    end;

    { -------------------------------------- }
    procedure ApplyQuery;
    begin
        try
            FPrevDataSetChanged := (ApplyMode = qaModify);
            if (ApplyMode = qaReplace) and (Language = qlSQL) then
                with FTmpQuery do begin
                    if not CanQueryDS(GetBaseDataSet) then
                        GenErrorResFmt(SCantSQLQuery, [GetBaseDataSet.ClassName]);
                    SessionName := TDBDataSet(FPrevDataSet).SessionName;
                    DatabaseName := TDBDataSet(FPrevDataSet).DatabaseName;
                    RequestLive := FPrevDataSet.CanModify and (qoSQLLive in FQBFOptions);
                    AssembleSQL(FTmpQuery, FPrevDataSet, '*', True);
                    if qoSQLCloneFields in FQBFOptions then
                        DUCloneDSFields(FPrevDataSet, FTmpQuery, qoSQLCloneCalcs in FQBFOptions);
                    if qoSQLCloneEvents in FQBFOptions then
                        DUCloneDSEvents(FPrevDataSet, FTmpQuery);
                    if not (qoSQLCloneCalcs in FQBFOptions) then
                        FTmpQuery.OnCalcFields := nil;
                    if not (qoSQLUseFilter in FQBFOptions) then
                        FTmpQuery.OnFilterRecord := nil;
                    if (qoSQLCachedUpdates in FQBFOptions) then begin
                        FTmpQuery.CachedUpdates := FPrevDataSet.CachedUpdates;
                        FTmpQuery.UpdateObject := FPrevDataSet.UpdateObject;
                    end
                    else begin
                        FTmpQuery.OnUpdateError := nil;
                        FTmpQuery.OnUpdateRecord := nil;
                    end;
                    DUPreviewSQL(qoSQLPreview in Options, FTmpQuery.SQL,
                        Self.Name + '.FTmpQuery.SQL');
                    Open;
                    Self.DataSource.DataSet := FTmpQuery;
                end
            else if ApplyMode = qaModify then begin
                if Language = qlFilter then begin
                    DUPreviewSQL(qoSQLPreview in Options, FSQL,
                        Self.Name + '.SQL');
                    AssembleFilter(FPrevDataSet);
                    DataSource.DataSet := FPrevDataSet;
                end
                else if Language = qlSQL then begin
                    AssembleSQL(TQuery(FPrevDataSet), FPrevDataSet, '*', True);
                    DUPreviewSQL(qoSQLPreview in Options, TQuery(FPrevDataSet).SQL,
                        Self.Name + '.' + FPrevDataSet.Name + '.SQL');
                    FPrevDataSet.Open;
                    DataSource.DataSet := FPrevDataSet;
                end
            end
            else if ApplyMode = qaNotify then begin
                DUPreviewSQL(qoSQLPreview in Options, FSQL,
                        Self.Name + '.SQL');
                DataSource.DataSet := FPrevDataSet;
            end;
        except
            if QueryBuilder = qiInplace then
                DataSource.DataSet := FQBFTable;
            RestoreState;
            raise;
        end;
    end;

begin
    if State = qsActive then
        Exit;
    if State = qsInactive then
        GenErrorRes(SQBFNotDefined);
    FQBFTable.CheckBrowseMode;
    try
        if Assigned(FOnBeforePost) then
            FOnBeforePost(Self);
        WDSuspend(False, '');
        try
            QBF2Query;
            ApplyQuery;
            if (QueryBuilder = qiInplace) and (State = qsEdit) then
                RestoreDetailSource;
            SetState(qsActive);
        finally
            WDResume;
        end;
        if Assigned(FOnAfterPost) then
            FOnAfterPost(Self);
    except
        on ExceptObject: EAbort do
            Cancel(False)
        else if IfPostError = qeError then
            raise
        else begin
            Application.HandleException(Self);
            case IfPostError of
            qeCancel: Cancel(False);
            qeEdit: if QueryBuilder = qiDialog then FLoopEdit := True else Edit;
            end;
        end;
    end;
end;

{ ---------------------------------------------------------------------- }
{ ---------------------------------------------------------------------- }
{ Locate }

procedure TNCQBF.Locate(reset: Boolean);
begin
    if not DataSetActive or (State = qsEdit) then begin
        warningDlg([LoadStr(SLocateBad)], -1);
        Exit;
    end;
    if reset or not CanNextLocate then begin
        if NCQBFLocateDlg = nil then
            NCQBFLocateDlg := TNCQBFLocateDlg.Create(Application);
        FOperationField := DUActiveField(DataSource.DataSet);
        FLocateMask := '';
        with NCQBFLocateDlg do begin
            Hide;
            QBF := Self;
            Show;
        end;
    end
    else
        DoLocate(False, 1);
end;

function TNCQBF.GetCanLocate: Boolean;
begin
    Result := DataSetActive and (FOperationField <> nil) and (FLocateMask <> '') and
        (DataSource.DataSet.FindField(FOperationField.FieldName) = FOperationField);
end;

procedure TNCQBF.DoLocate(restart: Boolean; dir: Integer);
var
    bmk: TBookmark;
    fnd: Boolean;

    function GetOperFieldText: String;
    begin
        with FOperationField do
            if (DataType in DUBlobFieldTypes) or (DataType in DUNumericFieldTypes) then
                Result := AsString
            else
                Result := DisplayText;
        Result := Trim(Result);
    end;

    function EOS: Boolean;
    begin
        with DataSource.DataSet do
            Result := ((dir > 0) and EOF) or ((dir < 0) and BOF);
    end;

    function DoDefaultLocate: Boolean;
    var
        i: Integer;
        ds: TDataSet;
        IndexDesc: IDXDesc;
        s: String;
    begin
        ds := DataSource.DataSet;
        i := Pos(ManyCharsMask, FLocateMask);
        if (ds is TTable) and
           (DbiGetIndexDesc(TTable(ds).Handle, 0, IndexDesc) = 0) and
                (IndexDesc.bCaseInsensitive = (qoNoCase in Options)) and
           (Pos(OneCharMask, FLocateMask) = 0) and
                ((i = 0) or (i = Length(FLocateMask)) and (Length(FLocateMask) > 1)) and
           (TTable(ds).IndexFields[DUCountField(TTable(ds).MasterFields) - 1] = FOperationField)
        then with TTable(ds) do begin
            Restart := Restart or not StrLike(GetOperFieldText, FLocateMask, qoNoCase in Options);
            if restart then
            try
                s := FLocateMask;
                if i <> 0 then
                    s := Copy(s, 1, Length(s) - 1);
                if dir < 0 then
                    s := Copy(s, 1, Length(s) - 1) + Chr(Ord(s[Length(s)]) + 1);
                FindNearest([s]);
                if dir < 0 then
                    Prior;
            finally
                Cancel;
            end
            else
                MoveBy(dir);
            Result := not EOS and StrLike(GetOperFieldText, FLocateMask, qoNoCase in Options);
        end
        else
        with ds do begin
            if restart then
                if dir > 0 then
                    First
                else
                    Last
            else
                MoveBy(dir);
            while not EOS and not StrLike(GetOperFieldText, FLocateMask, qoNoCase in Options) do
                MoveBy(dir);
            Result := not EOS;
        end;
    end;

begin
    if not CanNextLocate then begin
        warningDlg([LoadStr(SLocateBad)], -1);
        Exit;
    end;
    with DataSource.DataSet do begin
        bmk := GetBookmark;
        DisableControls;
        try
            WDSuspend(False, LoadStr(SLocateGo));
            try
                if Assigned(FOnLocate) then
                    FOnLocate(Self, restart, dir, fnd)
                else
                    fnd := DoDefaultLocate;
            finally
                WDResume;
            end;
            if not fnd then begin
                GotoBookmark(bmk);
                warningDlg([LoadStr(SLocateNotFound)], -1);
            end;
        finally
            FreeBookmark(bmk);
            EnableControls;
        end;
    end;
    StateChanged;
end;

{ ---------------------------------------------------------------------- }
{ ---------------------------------------------------------------------- }
{ Calculate }

procedure TNCQBF.Calculate;
begin
    if not DataSetActive or (State = qsEdit) then begin
        warningDlg([LoadStr(SCalculateBad)], -1);
        Exit;
    end;
    if NCQBFCalculateDlg = nil then
        NCQBFCalculateDlg := TNCQBFCalculateDlg.Create(Application);
    with NCQBFCalculateDlg do begin
        DataSource := Self.DataSource;
        Field := DUActiveField(DataSource.DataSet);
        ShowHint := Self.ShowHint;
        LockDistinct := qoDistinct in Options;
        if (ShowModal = mrOK) then begin
            if (Field = nil) or (Operation - [coDistinct] = []) then begin
                warningDlg([LoadStr(SCalculateBad)], -1);
                Exit;
            end;
            FOperationField := Field;
            FCalculateOperation := Operation;
            InformationDlg([Format(LoadStr(SCalculateResult),
                [NCQBFCalculateDlg.SelOperCB.Text, FOperationField.DisplayLabel,
                 DoCalculate])], -1);
        end;
    end;
end;

function TNCQBF.SQLSelect4Calc(AExpr: String): String;
begin
    if coSUM in FCalculateOperation then
        Result := 'SUM('
    else if coCOUNT in FCalculateOperation then
        Result := 'COUNT('
    else if coAVERAGE in FCalculateOperation then
        Result := 'AVG('
    else if coMIN in FCalculateOperation then
        Result := 'MIN('
    else if coMAX in FCalculateOperation then
        Result := 'MAX(';
    if (coDISTINCT in FCalculateOperation) or (qoDistinct in Options) then
        Result := Result + 'DISTINCT ';
    Result := Result + AExpr + ')';
end;

function TNCQBF.DoCalculate: String;
var
    bmk: TBookmark;
    prevLT: TSQLLookupTemplate;

    function DoIterateCalculate: String;
    var
        V: Variant;
        Cnt: LongInt;
        tp: Integer;
    begin
        Cnt := 0;
        V := Null;
        Result := '';
        tp := varEmpty;
        with DataSource.DataSet do begin
            First;
            while not EOF do begin
                if not FOperationField.IsNull then begin
                    Inc(Cnt);
                    tp := VarType(FOperationField.Value);
                    if coSUM in FCalculateOperation then begin
                        if VarIsNull(V) then
                            V := FOperationField.Value
                        else
                            V := V + FOperationField.Value
                    end
                    else if coMIN in FCalculateOperation then begin
                        if VarIsNull(V) or (V > FOperationField.Value) then
                            V := FOperationField.Value
                    end
                    else if coMAX in FCalculateOperation then begin
                        if VarIsNull(V) or (V < FOperationField.Value) then
                            V := FOperationField.Value
                    end
                    else if coAVERAGE in FCalculateOperation then begin
                        if VarIsNull(V) then
                            V := FOperationField.Value
                        else
                            V := V + FOperationField.Value;
                    end;
                end;
                Next;
            end;
        end;
        if coCOUNT in FCalculateOperation then
            Result := IntToStr(Cnt)
        else if not VarIsNull(V) then begin
            if coAVERAGE in FCalculateOperation then
                V := V / Cnt;
            Result := VarAsType(VarAsType(V, tp), varString);
        end;
    end;

    function DoSQLCalculate: String;
    var
        FieldPreSQL, FieldNameSQL, FieldPostSQL, FieldFromSQL, dsFrom,
            dsWhere, dsOrderBy: String;
        FieldDataType: TFieldType;
        FieldValueRetran: TValueRetranEvent;
        wasWhere: Boolean;
        dsMasterDS: TDataSource;
        dbDs: TDBDataSet;
    begin
        with TQuery.Create(nil) do
        try
            dbDs := TDBDataSet(Self.DataSource.DataSet);
            SessionName := dbDs.SessionName;
            DatabaseName := dbDs.DatabaseName;
            Unidirectional := True;
            BuildFieldSQL(FOperationField, FieldPreSQL, FieldNameSQL,
                FieldPostSQL, FieldFromSQL, FieldDataType, FieldValueRetran);
            BuildDataSetSQL(GetBaseDataSet, SQLAlias, dsFrom, dsWhere,
                dsOrderBy, dsMasterDS, Params);
            SQL.BeginUpdate;
            try
                SQL.Add('select ' + SQLSelect4Calc(FieldNameSQL));
                SQL.Add('from ' + Strs2Str([dsFrom, FSQLFrom, FieldFromSQL], True, '', ',', ''));
                wasWhere := False;
                if dsWhere <> '' then
                    DUContWhere(SQL, wasWhere, dsWhere);
                if FSQL.Count <> 0 then
                    if Language = qlSQL then
                        DUContWhere(SQL, wasWhere, FSQL.Text)
                    else { if (ApplyMode = qaNotify) then }
                        DUContWhere(SQL, wasWhere, DUFilter2SQL(FSQL.Text, SQLAlias));
                if (FieldPreSQL <> '') or (FieldPostSQL <> '') then
                    DUContWhere(SQL, wasWhere, '(' + FieldPreSQL + '(1 = 1)' + FieldPostSQL + ')');
                DataSource := dsMasterDS;
            finally
                SQL.EndUpdate;
            end;
            Open;
            Result := Fields[0].AsString;
        finally
            Free;
        end;
    end;

begin
    with DataSource.DataSet do begin
        WDSuspend(False, LoadStr(SCalculateGo));
        bmk := GetBookmark;
        DisableControls;
        prevLT := SQLLookupTemplate;
        SQLLookupTemplate := ltJoin;
        try
            if Assigned(FOnCalculate) then
                FOnCalculate(Self, Result)         // can build simple query ???
            else if CanQueryDS(GetBaseDataSet) and (tkGR in CanQueryField(FOperationField)) and
                    not Assigned(GetBaseDataSet.OnFilterRecord) then
                Result := DoSQLCalculate
            else
                Result := DoIterateCalculate;
        finally
            SQLLookupTemplate := prevLT;
            EnableControls;
            GotoBookmark(bmk);
            FreeBookmark(bmk);
            WDResume;
        end;
    end;
end;

{ ---------------------------------------------------------------------- }
{ ---------------------------------------------------------------------- }

destructor TQBFLink.Destroy;
begin
    SetQBF(nil);
    inherited Destroy;
end;

function TQBFLink.GetState: TQBFState;
begin
    Result := qsInactive;
    if FQBF <> nil then
        Result := FQBF.State;
end;

procedure TQBFLink.SetQBF(AQBF: TNCQBF);
begin
    if FQBF <> nil then
        FQBF.RemoveLink(Self);
    if AQBF <> nil then
        AQBF.AddLink(Self);
end;

procedure TQBFLink.StateChanged;
begin
end;

{ ---------------------------------------------------------------------- }
{*********************************************************************** }
{                IT IS A TRANSLATOR ! DANGER !                           }
{*********************************************************************** }
{ ---------------------------------------------------------------------- }

constructor TNCQBFTranslator.Create;
begin
    inherited Create;
    FLang := qlSQL;
    FOptions := [toNoCase];
end;

procedure TNCQBFTranslator.Translate(AFieldLabel, AFieldName: String; AFieldType: TFieldType;
    AEnTokens: TQBFTokens; ACond: String; ARetranEvent: TValueRetranEvent);
begin
    FItemQBF := ACond;
    FItemFieldType := AFieldType;
    FItemFieldLabel := AFieldLabel;
    FItemFieldName := AFieldName;
    FEnTokens := AEnTokens;
    FItemSort := 0;
    FItemSQL := '';
    FParenthesisLevel := 0;
    FParsePos := 1;
    FValueRetranEvent := ARetranEvent;
    Term1;
    if FParenthesisLevel <> 0 then
        GenErrorRes(SNoRPar);
end;

{ ---------------------------------------------------------------------- }
{ ---------------------------------------------------------------------- }
{ Tokenizer }

procedure TNCQBFTranslator.ReadToken;
type
    TCharSet = set of char;
const
    SpecCharsWD: TCharSet = ['(', ')', '~', '|', '&', '[', '@', '=', '<', '>', '"',
        '''', '#', '!'];
    DirCharsWD: TCharSet = ['(', ')', '~', '|', '&', '[', '@', '#', '!'];
    SpecCharsND: TCharSet = ['(', ')', '~', '|', '&', '@', '=', '<', '>', '"',
        '''', '#', '!'];
    DirCharsND: TCharSet = ['(', ')', '~', '|', '&', '@', '#', '!'];
    CmpChars: TCharSet = ['=', '<', '>'];
    DelimChars: TCharSet = ['"', ''''];
    MacroDelim: TCharSet = [' ', '.', '@'];
var
    waitDelim: Char;
    SpecChars, DirChars: TCharSet;

    procedure SkipBlanks;
    begin
        while (FParsePos <= length(FItemQBF)) and (FItemQBF[FParsePos] = ' ') do inc(FParsePos);
    end;

    procedure ExtractToken(Chars: TCharSet; whileNot: Boolean);
    begin
        while (FParsePos <= length(FItemQBF)) and (whileNot xor (FItemQBF[FParsePos] in Chars)) do begin
            FTokenStr := FTokenStr + FItemQBF[FParsePos];
            inc(FParsePos)
        end;
    end;

    procedure ExpandMacro(var s: String);
    var
        found: Boolean;
        prevS: String;
    begin
        s := AnsiUpperCase(s);
        if s = 'X' then
            s := FItemFieldName
        else if s = 'DATE' then
            s := DateToStr(Date)
        else begin
            prevS := s;
            found := False;
            if (FMacros <> nil) then begin
                s := FMacros.Values[s];
                found := s <> '';
            end;
            if not found then
                GenErrorResFmt(SMacNotFound, [prevS]);
        end;
    end;

    procedure ExpandStr(var s: String);
    var
        i, j: Integer;
        s1: String;
    begin
        i := 1;
        while i <= Length(s) do begin
            if s[i] = '@' then begin
                j := i + 1;
                while (j <= Length(s)) and not (s[j] in MacroDelim) do
                    Inc(j);
                s1 := Copy(s, i + 1, j - i - 1);
                ExpandMacro(s1);
                if (j > Length(s)) or (s[j] <> ' ') then
                    Inc(j);
                s := Stuff(s, i, j - i, s1);
                i := i + Length(s1);
            end
            else
                Inc(i);
        end;
    end;

begin
    if toNoDirect in FOptions then begin
        SpecChars := SpecCharsND;
        DirChars := DirCharsND;
    end
    else begin
        SpecChars := SpecCharsWD;
        DirChars := DirCharsWD;
    end;
    SkipBlanks;
    if FParsePos > length(FItemQBF) then begin
        FToken := tkEND;
        FTokenStr := '<END>';
        Exit;
    end;
    FTokenStr := '';
    if FItemQBF[FParsePos] in CmpChars then begin
        ExtractToken(CmpChars, False);
        if FTokenStr = '=' then
            FToken := tkEQ
        else if FTokenStr = '>' then
            FToken := tkGR
        else if FTokenStr = '<' then
            FToken := tkLO
        else if FTokenStr = '>=' then
            FToken := tkGE
        else if FTokenStr = '<=' then
            FToken := tkLE
        else if FTokenStr = '<>' then
            FToken := tkNOTEQ
        else
            GenErrorResFmt(SBadOperator, [FTokenStr]);
    end
    else if FItemQBF[FParsePos] in DirChars then begin
        FTokenStr := FItemQBF[FParsePos];
        inc(FParsePos);
        if FTokenStr = '~' then
            FToken := tkBETWEEN
        else if FTokenStr = '|' then
            FToken := tkOR
        else if FTokenStr = '&' then
            FToken := tkAND
        else if FTokenStr = '(' then
            FToken := tkLPAR
        else if FTokenStr = ')' then
            FToken := tkRPAR
        else if FTokenStr = '[' then begin
            FToken := tkDIRECT;
            FTokenStr := '';
            ExtractToken([']'], True);
            ExpandStr(FTokenStr);
            inc(FParsePos);
        end
        else if FTokenStr = '@' then begin
            FToken := tkVALUE;
            FTokenStr := '';
            ExtractToken(SpecChars + MacroDelim, True);
            ExpandMacro(FTokenStr);
        end
        else if FTokenStr = '#' then begin
            ExtractToken(['-', '0'..'9'], False);
            if FTokenStr[2] = '-' then
                FToken := tkSORTDESC
            else
                FToken := tkSORTASC;
        end
        else if FTokenStr = '!' then
            FToken := tkNOT;
    end
    else begin
        if (FItemQBF[FParsePos] in DelimChars) then begin
            waitDelim := FItemQBF[FParsePos];
            inc(FParsePos);
            ExtractToken([waitDelim], True);
            inc(FParsePos);
        end
        else begin
            ExtractToken(SpecChars, True);
            FTokenStr := Trim(FTokenStr);
        end;
        if (Pos('%', FTokenStr) <> 0) or (Pos('_', FTokenStr) <> 0) then
            FToken := tkMASK
        else
            FToken := tkVALUE;
    end;
end;

procedure TNCQBFTranslator.ReadValue;
begin
    ReadToken;
    if not (FToken in [tkVALUE, tkDIRECT]) then
        GenErrorResFmt(SValueExpected, [FTokenStr]);
end;

procedure TNCQBFTranslator.TokenAllowed(t: TQBFToken);
begin
    if not (t in FEnTokens) then
        GenErrorResFmt(SOperationDisabled, [FTokenStr, FItemFieldLabel]);
end;

function TNCQBFTranslator.ReadTokenIF(needToken: TQBFToken; MastUndo: Boolean): Boolean;
var
    saveParsePos: Integer;
    saveToken: TQBFToken;
    saveTokenStr: String;
begin
    saveParsePos := FParsePos;
    saveToken := FToken;
    saveTokenStr := FTokenStr;
    ReadToken;
    Result := (FToken = needToken);
    if not Result or MastUndo then begin
        FToken := saveToken;
        FTokenStr := saveTokenStr;
        FParsePos := saveParsePos;
    end;
end;

{ ---------------------------------------------------------------------- }
{ ---------------------------------------------------------------------- }
{ Parser }

procedure TNCQBFTranslator.Term1;
begin
    repeat
        Term0;
        ReadToken;
        TokenAllowed(FToken);
        case FToken of
            tkOR:
                begin
                    WriteStr(' OR ');
                    Term1;
                end;
            tkAND:
                begin
                    WriteStr(' AND ');
                    Term1;
                end;
            tkRPAR, tkEND:;
            else
                GenErrorResFmt(SBadExp, [FTokenStr]);
        end;
    until (FToken in [tkEND, tkRPAR]);
end;


procedure TNCQBFTranslator.Term0;
var
    i: Integer;
begin
    ReadToken;
    if not (FToken in [tkEQ, tkNOTEQ, tkDIRECT, tkVALUE, tkMASK]) then
        TokenAllowed(FToken);
    case FToken of
        tkEQ:
            if ReadTokenIF(tkVALUE, False) then begin
                TokenAllowed(tkEQ);
                WriteBinOp(' = ');
            end
            else begin
                TokenAllowed(tkNULL);
                if FLang = qlFilter then begin
                    WriteFieldName;
                    WriteStr(' = NULL ');
                end
                else begin
                    WriteStr(FItemFieldName);
                    WriteStr(' IS NULL ');
                end;
            end;
        tkGR:
            begin
                ReadValue;
                WriteBinOp(' < ');
            end;
        tkLO:
            begin
                ReadValue;
                WriteBinOp(' > ');
            end;
        tkGE:
            begin
                ReadValue;
                WriteBinOp(' <= ');
            end;
        tkLE:
            begin
                ReadValue;
                WriteBinOp(' >= ');
            end;
        tkNOTEQ:
            if ReadTokenIF(tkVALUE, False) then begin
                TokenAllowed(tkNOTEQ);
                WriteStr(' NOT (');
                WriteBinOp(' = ');
                WriteStr(')');
            end
            else begin
                TokenAllowed(tkNOTNULL);
                if FLang = qlFilter then begin
                    WriteStr(' NOT (');
                    WriteFieldName;
                    WriteStr(' = NULL)');
                end
                else begin
                    WriteStr(FItemFieldName);
                    WriteStr(' IS NOT NULL ');
                end;
            end;
        tkDIRECT, tkVALUE:
            if (FToken = tkDIRECT) and not (toDirectAsValue in FOptions) then begin
                TokenAllowed(tkDIRECT);
                WriteStr(FTokenStr)
            end
            else begin
                if not ReadTokenIF(tkBETWEEN, True) then begin
                    TokenAllowed(tkEQ);
                    WriteBinOp(' = ')
                end
                else begin
                    TokenAllowed(tkBETWEEN);
                    if Language = qlSQL then begin
                        WriteFieldName;
                        WriteStr(' BETWEEN ');
                        writeTokenStr;
                        WriteStr(' AND ');
                        ReadToken;
                        ReadValue;
                        WriteTokenStr;
                    end
                    else begin
                        WriteStr('(');
                        WriteTokenStr;
                        WriteStr(' <=');
                        WriteFieldName;
                        WriteStr(' AND ');
                        ReadToken;
                        ReadValue;
                        WriteTokenStr;
                        WriteStr(' >= ');
                        WriteFieldName;
                        WriteStr(')');
                    end;
                end;
            end;
        tkMASK:
            if Language = qlSQL then begin
                WriteFieldName;
                FItemSQL := FItemSQL + ' LIKE ';
                WriteTokenStr;
            end
            else begin
                if (Pos('_', FTokenStr) <> 0) or (Pos('%', FTokenStr) <> Length(FTokenStr)) then
                    TokenAllowed(tkIMASK);
                FTokenStr := Copy(FTokenStr, 1, Length(FTokenStr) - 1);
                WriteStr('(');
                WriteTokenStr;
                WriteStr(' <=');
                WriteFieldName;
                WriteStr(' AND ');
                i := Length(FTokenStr);
                FTokenStr[i] := Chr(Ord(FTokenStr[i]) + 1);
                WriteTokenStr;
                WriteStr(' > ');
                WriteFieldName;
                WriteStr(')');
            end;
        tkSORTASC, tkSORTDESC:
            begin
                FItemSort := StrToInt(Copy(FTokenStr, 2, Length(FTokenStr)));
                Term0;
            end;
        tkNOT:
            begin
                WriteStr(' NOT ');
                Term0;
            end;
        tkLPAR:
            begin
                Inc(FParenthesisLevel);
                WriteStr('(');
                Term1;
                if FToken <> tkRPAR then
                    GenErrorResFmt(SRParExpected, [FTokenStr]);
                WriteStr(')');
                Dec(FParenthesisLevel);
            end;
        tkRPAR, tkEND:;
        else
            GenErrorResFmt(SBadExp, [FTokenStr]);
    end;
end;

procedure TNCQBFTranslator.WriteStr(const s: String);
begin
    FItemSQL := FItemSQL + s;
end;

procedure TNCQBFTranslator.WriteBinOp(const op: String);
begin
    WriteTokenStr;
    FItemSQL := FItemSQL + op;
    WriteFieldName;
end;

procedure TNCQBFTranslator.WriteFieldName;
var
    s: String;
begin
    if FLang = qlFilter then
        FItemSQL := FItemSQL + '[' + FItemFieldName + ']'
    else begin
        s := FItemFieldName;
        if (FToken = tkMASK) and not SQLServer.Can(scAutoCast) then
            s := SQLServer.CastToStr(s);
        if (toNoCase in FOptions) and (FItemFieldType in [ftString, ftMemo]) then
            s := SQLServer.UpperCase(s);
        FItemSQL := FItemSQL + s;
    end;
end;

procedure TNCQBFTranslator.WriteTokenStr;
var
    s: String;
    v: Variant;
begin
    s := FTokenStr;
    if Assigned(FValueRetranEvent) then begin
        v := FValueRetranEvent(s);
        if VarIsNull(v) then begin
            FItemSQL := FItemSQL + ' NULL ';
            s := '';
        end
        else
            s := Var2Text(v);
    end;
    if s <> '' then begin
        if (toNoCase in FOptions) and (FToken <> tkDIRECT) then
            s := AnsiUpperCase(s);
        if (FItemFieldType in [ftDate, ftTime, ftDateTime]) and (FToken = tkVALUE) or
           (FItemFieldType in [ftString, ftMemo]) or
           (FToken = tkMASK) then
            s := '''' + s + '''';
        FItemSQL := FItemSQL + s;
    end;
end;

end.
