{***********************************************}
{File:      NCQBFExp.PAS                        }
{Revision:  2.02 / 26.12.1999                   }
{Comment:   QBF condition editor dialog         }
{Copyright: (c) 1997-1999, Nica-Com Ltd.        }
{E-Mail:    diman@ncom.nnov.su                  }
{***********************************************}

unit NCQBFExp;

interface

uses Windows, Classes, Graphics, Forms, Controls, Buttons, StdCtrls, Mask,
     ExtCtrls, DBTables, DB, NCQBF, Grids, NCDbREdt, NCCalndr, NCBrowse,
     NCCalc {$IFDEF NCQBFLOOKUP}, NCBrArr {$ENDIF};

type
    TNCQBFCondDesc = class
    private
        FField: TField;
        FQBF: TNCQBF;
        FOper: TQBFToken;
        FArg1: String;
        FArg2: String;
        FArg3: TStringList;
        FNOT: Boolean;
        FEnOpers: TQBFTokens;
        FSortOrder: Integer;
        function Ind2OperDesc(i: Integer): Integer;
        procedure QBF2Cond(s: String);
        function Cond2QBF: String;
        function UserString: String;
    public
        constructor Create(qbf1: TNCQBF; fld1: TField);
        destructor Destroy; override;
        procedure Clear;
        function Edit: Boolean;

        property Field: TField read FField;
        property Order: Integer read FSortOrder write FSortOrder;
        property Condition: String read Cond2QBF write QBF2Cond;
        property Description: String read UserString;
        property QBF: TNCQBF read FQBF;
    end;

    TExpType = (etNone, etSimple, etBetween, etList);

    TNCQBFConditionDlg = class(TForm)
        Panel2: TPanel;
        SelectOperCB: TComboBox;
        Val1Edit: TNCDBReferEdit;
        Val2Edit: TNCDBReferEdit;
        Val3LB: TListBox;
        Panel3: TPanel;
        AddValBtn: TSpeedButton;
        DelValBtn: TSpeedButton;
        FieldNameLbl: TLabel;
        Label1: TLabel;
        EditValBtn: TSpeedButton;
        Panel1: TPanel;
        OKBtn: TBitBtn;
        CancelBtn: TBitBtn;
        HelpBtn: TBitBtn;
        NOTChb: TCheckBox;
        clcltr: TNCCalculator;
        clndr: TNCCalendar;
        procedure SelectOperCBClick(Sender: TObject);
        procedure AddValBtnClick(Sender: TObject);
        procedure DelValBtnClick(Sender: TObject);
        procedure EditValBtnClick(Sender: TObject);
        procedure Val3LBKeyDown(Sender: TObject; var Key: Word;
            Shift: TShiftState);
        procedure NOTChbClick(Sender: TObject);
        procedure FormShow(Sender: TObject);
    private
        { Private declarations }
        Desc: TNCQBFCondDesc;
        ExpType: TExpType;
{$IFDEF NCQBFLOOKUP}
        brwsr: TNCArrBrowser;
{$ENDIF}
        procedure SetupDialog(noCompare: Boolean);
    public
        { Public declarations }
        constructor Create(AOwner: TComponent); override;
        function Execute(ADesc: TNCQBFCondDesc): Boolean;
    end;

var
  NCQBFConditionDlg: TNCQBFConditionDlg;

{ ---------------------------------------------------------------- }
{ ---------------------------------------------------------------- }

implementation

{$R *.DFM}

Uses SysUtils, NCUtil, NCUIUtil, Dialogs, NCStrs, consts, NCRefFld;

type
    TOperDesc = record
        token: TQBFToken;
        expType: TExpType;
        usrFmt: String;
        operName: String;
    end;

var
    OperDesc: array[-1..13] of TOperDesc = (
        (token: tkEND;      expType: etNone;    usrFmt: ''; operName: ''),
        (token: tkNULL;     expType: etNone;    usrFmt: ''; operName: ''),
        (token: tkNOTNULL;  expType: etNone;    usrFmt: ''; operName: ''),
        (token: tkEQ;       expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkNOTEQ;    expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkGR;       expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkGE;       expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkLO;       expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkLE;       expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkSMASK;    expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkEMASK;    expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkIMASK;    expType: etSimple;  usrFmt: ''; operName: ''),
        (token: tkBETWEEN;  expType: etBetween; usrFmt: ''; operName: ''),
        (token: tkOR;       expType: etLIST;    usrFmt: ''; operName: ''),
        (token: tkDIRECT;   expType: etSimple;  usrFmt: ''; operName: ''));
    OperDescLoaded: Boolean;
    NOTChbCapts: array[Boolean] of String;
    NOUsrFmt: String;

Const
    DlgHArr: array[TExpType] of Integer = (156, 190, 226, 345);

{ ---------------------------------------------------------------------------- }
{ ---------------------------------------------------------------------------- }

constructor TNCQBFCondDesc.Create(qbf1: TNCQBF; fld1: TField);
var
    i: Integer;
begin
    inherited Create;
    if not OperDescLoaded then begin
        for i := 0 to High(OperDesc) do begin
            OperDesc[i].usrFmt := LoadStr(SUserStrBase + i);
            OperDesc[i].operName := LoadStr(SSelOperBase + i);
        end;
        NOTChbCapts[False] := {$IFDEF D2}LoadStr({$ENDIF} SYesButton {$IFDEF D2}){$ENDIF};
        NOTChbCapts[True] := {$IFDEF D2}LoadStr({$ENDIF} SNoButton {$IFDEF D2}){$ENDIF};
        NOUsrFmt := LoadStr(SNOUsrStr);
        OperDescLoaded := True;
    end;
    FQBF := qbf1;
    FField := fld1;
    FOper := tkEND;
    FArg1 := '';
    FArg2 := '';
    FArg3 := TStringList.Create;
    FSortOrder := 0;
    FEnOpers := FQBF.CanQueryField(FField);
end;

destructor TNCQBFCondDesc.Destroy;
begin
    FArg3.Free;
    inherited Destroy;
end;

procedure TNCQBFCondDesc.Clear;
begin
    FOper := tkEND;
    FSortOrder := 0;
    FArg1 := '';
    FArg2 := '';
    FArg3.Clear;
    FNOT := False;
end;

procedure TNCQBFCondDesc.QBF2Cond(s: String);
type
    CharSet = set of char;
var
    i, i1, prevI: Integer;
    f1, f2, l1: String;

    procedure SkipChr(what: CharSet);
    begin
        while (i <= length(s)) and (s[i] in what) do
            inc(i);
    end;

    procedure DoOrderBy;
    begin
        FSortOrder := 0;
        if (Length(s) >= 1) and (s[1] = '#') then begin
            i := 2;
            SkipChr(['-', '0'..'9']);
            Val(Copy(s, 2, i - 2), FSortOrder, i1);
            SkipChr([' ']);
            s := Copy(s, i, Length(s));
        end;
    end;

    procedure DoNOT;
    begin
        FNOT := False;
        if (Length(s) >= 1) and (s[1] = '!') then begin
            i := 2;
            SkipChr([' ']);
            if (s[i] = '(') and (s[Length(s)] = ')') then
                s := Copy(s, i + 1, Length(s) - i - 1);
            FNOT := True;
        end;
    end;

    procedure DoSimple(ind: TQBFToken; from: Integer);
    begin
        FOper := ind;
        FArg1 := Copy(s, from, Length(s));
    end;

    procedure DoDirect(ind: TQBFToken);
    begin
        FOper := ind;
        FArg1 := Copy(s, 2, Length(s) - 2);
    end;

    procedure DoMask(ind: TQBFToken);
    begin
        FOper := ind;
        i := 1;
        while (s[i] = '%') and (i <= Length(s)) do
            Inc(i);
        i1 := Length(s);
        while (s[i1] = '%') and (i1 >= 1) do
            Dec(i1);
        FArg1 := Copy(s, i, i1 - i + 1);
    end;

    procedure DoBetween(ind: TQBFToken);
    begin
        FOper := ind;
        i := Pos('~', s);
        FArg1 := Copy(s, 1, i - 2);
        FArg2 := Copy(s, i + 2, Length(s));
    end;

    procedure DoList(indONE, indMANY: TQBFToken);
    begin
        prevI := 1;
        i := 1;
        while i <= Length(s) do begin
            if s[i] = '|' then begin
                FArg3.Add(Copy(s, prevI, i - 1 - prevI));
                prevI := i + 2;
                i := i + 2;
            end;
            Inc(i);
        end;
        if FArg3.Count = 0 then begin
            FArg1 := Copy(s, prevI, Length(s));
            FOper := indONE;
        end
        else begin
            FArg3.Add(Copy(s, prevI, Length(s)));
            FOper := indMANY;
        end;
    end;

begin
    Clear;
    if s = '' then
        Exit;
    DoOrderBy;
    DoNOT;
    if s = '' then
        Exit;
    f1 := s[1];
    f2 := Copy(s, 1, 2);
    l1 := s[Length(s)];
    if Pos('|', s) <> 0 then
        DoList(tkEQ, tkOR)
    else if s = '=' then
        FOper := tkNULL
    else if s = '<>' then
        FOper := tkNOTNULL
    else if f1 = '=' then
        DoSimple(tkEQ, 3)
    else if f2 = '<>' then
        DoSimple(tkNOTEQ, 4)
    else if f2 = '>=' then
        DoSimple(tkGE, 4)
    else if f1 = '>' then
        DoSimple(tkGR, 3)
    else if f2 = '<=' then
        DoSimple(tkLE, 4)
    else if f1 = '<' then
        DoSimple(tkLO, 3)
    else if (l1 = '%') and (f1 <> '%') then
        DoMask(tkSMASK)
    else if (l1 <> '%') and (f1 = '%') then
        DoMask(tkEMASK)
    else if (l1 = '%') and (f1 = '%') then
        DoMask(tkIMASK)
    else if (l1 = '[') and (f1 = ']') then
        DoDirect(tkDIRECT)
    else if Pos('~', s) <> 0 then
        DoBetween(tkBETWEEN)
    else
        DoList(tkEQ, tkOR);
end;

function TNCQBFCondDesc.Cond2QBF: String;
var
    i: Integer;
begin
    case FOper of
    tkNULL: Result := '=';
    tkNOTNULL: Result := '<>';
    tkEQ: Result := '= ' + FArg1;
    tkNOTEQ: Result := '<> ' + FArg1;
    tkGR: Result := '> ' + FArg1;
    tkGE: Result := '>= ' + FArg1;
    tkLO: Result := '< ' + FArg1;
    tkLE: Result := '<= ' + FArg1;
    tkSMASK: Result := FArg1 + '%';
    tkEMASK: Result := '%' + FArg1;
    tkIMASK: Result := '%' + FArg1 + '%';
    tkBETWEEN: Result := FArg1 + ' ~ ' + FArg2;
    tkDIRECT: Result := '[' + FArg1 + ']';
    tkOR:
        begin
            Result := '';
            for i := 0 to FArg3.Count - 1 do begin
                Result := Result + FArg3[i];
                if i < FArg3.Count - 1 then
                    Result := Result + ' | ';
            end;
        end;
    tkEND: Result := '';
    end;
    if FNOT and (Result <> '') then
        Result := '! (' + Result + ')';
    if FSortOrder <> 0 then
        Result := '#' + IntToStr(FSortOrder) + ' ' + Result;
end;

function TNCQBFCondDesc.UserString: String;
var
    j, i: Integer;
begin
    if FOper = tkEND then
        Result := ''
    else begin
        for j := 0 to High(OperDesc) do
            if OperDesc[j].token = FOper then begin
                Result := OperDesc[j].usrFmt;
                Break;
            end;
        if FOper <> tkOR then
            Result := Format(Result, [FArg1, FArg2])
        else begin
            for i := 0 to FArg3.Count - 1 do begin
                Result := Result + '"' + FArg3[i] + '"';
                if i < FArg3.Count - 1 then
                    Result := Result + ', ';
            end;
        end;
        if FNOT and (Result <> '') then
            Result := Format(NOUsrFmt, [Result]);
    end;
    if Assigned(FQBF.OnRenderCond) then
        FQBF.OnRenderCond(Self, Result);
end;

function TNCQBFCondDesc.Edit: Boolean;
var
    defEdit: Boolean;
begin
    defEdit := True;
    Result := False;
    if Assigned(FQBF.OnEditCond) then
        FQBF.OnEditCond(Self, defEdit, Result);
    if defEdit then begin
        if NCQBFConditionDlg = nil then
            NCQBFConditionDlg := TNCQBFConditionDlg.Create(Application);
        Result := NCQBFConditionDlg.Execute(Self);
    end;
end;

function TNCQBFCondDesc.Ind2OperDesc(i: Integer): Integer;
begin
    for Result := 0 to High(OperDesc) do
        if OperDesc[Result].token in FEnOpers then begin
            Dec(i);
            if i < 0 then
                Exit;
        end;
    Result := -1;
end;

{ ---------------------------------------------------------------- }
{ ---------------------------------------------------------------- }

constructor TNCQBFConditionDlg.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    ExpType := etNone;
    IUStr2Font(LoadStr(SDefaultFont), Font);
{$IFDEF NCQBFLOOKUP}
    brwsr := TNCArrBrowser.Create(Self);
{$ENDIF}    
end;

procedure TNCQBFConditionDlg.SetupDialog(noCompare: Boolean);
var
    NewExpType: TExpType;

    procedure ShowItems(vis: Boolean);
    begin
        case ExpType of
        etList:
            begin
                Val3LB.Visible := vis;
                AddValBtn.Visible := vis;
                DelValBtn.Visible := vis;
                EditValBtn.Visible := vis;
            end;
        etSimple:
            begin
                Val1Edit.Visible := vis;
            end;
        etBetween:
            begin
                Val1Edit.Visible := vis;
                Val2Edit.Visible := vis;
            end;
        end;
    end;

begin
    NewExpType := OperDesc[Desc.Ind2OperDesc(SelectOperCB.ItemIndex)].ExpType;
    if (NewExpType <> ExpType) or noCompare then begin
        ShowItems(False);
        ExpType := NewExpType;
        Height := DlgHArr[NewExpType];
        ShowItems(True);
    end;
end;

procedure TNCQBFConditionDlg.SelectOperCBClick(Sender: TObject);
begin
    SetupDialog(False);
end;

function TNCQBFConditionDlg.Execute(ADesc: TNCQBFCondDesc): Boolean;
var
    newIndex, i: Integer;

    procedure SetupValEdit(ValEdit: TNCDBReferEdit);
    begin
        ValEdit.BeginUpdate;
        try
            if RFCheckRefer(ADesc.FField) then begin
                ValEdit.RefBrowser := TNCReferField(ADesc.FField).RefBrowser;
                ValEdit.RefDisplayFormat := TNCReferField(ADesc.FField).RefDisplayFormat;
            end
{$IFDEF NCQBFLOOKUP}
            else if ADesc.FField.FieldKind = fkLookup then begin
                ValEdit.RefBrowser := brwsr;
                ValEdit.RefDisplayFormat := '';
                brwsr.LookupKind := lkLookupField;
                brwsr.LookupField := ADesc.FField;
            end
{$ENDIF}
            else if ADesc.Field.DataType in [ftSmallint, ftInteger, ftWord, ftFloat,
                                             ftCurrency, ftBCD] then begin
                ValEdit.RefBrowser := clcltr;
                ValEdit.RefDisplayFormat := '';
            end
            else if ADesc.Field.DataType in [ftDate, ftTime, ftDateTime] then begin
                ValEdit.RefBrowser := clndr;
                ValEdit.RefDisplayFormat := '';
            end
            else begin
                ValEdit.RefBrowser := nil;
                ValEdit.RefDisplayFormat := '';
            end;
        finally
            ValEdit.EndUpdate;
        end;
    end;

begin
    Desc := ADesc;
    newIndex := -1;
    with SelectOperCB do begin
        Items.Clear;
        for i := 0 to High(OperDesc) do begin
            if OperDesc[i].token in Desc.FEnOpers then
                Items.Add(OperDesc[i].OperName);
            if (OperDesc[i].token = Desc.FOper) or
               (Desc.FOper = tkEND) and (OperDesc[i].token in Desc.FEnOpers) and (
                    (OperDesc[i].token = tkEQ) and
                    (Desc.FField.DataType in [ftString, ftSmallint, ftInteger, ftWord, ftBoolean,
                        ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime, ftAutoInc]) or
                    (OperDesc[i].token = tkIMASK) and
                    (Desc.FField.DataType in [ftString])
               ) then
                newIndex := Items.Count - 1;
        end;
    end;
    SelectOperCB.ItemIndex := newIndex;
    with desc do begin
        SetupValEdit(Val1Edit);
        Val1Edit.DisplayValue := FArg1;
        SetupValEdit(Val2Edit);
        Val2Edit.DisplayValue := FArg2;
        Val3LB.Items.Assign(FArg3);
        NOTChb.Checked := not FNOT;
        NOTChbClick(nil);
    end;
    with FieldNameLbl do begin
        Caption := desc.FField.DisplayLabel;
        Hint := desc.FField.FieldName;
    end;
    SetupDialog(True);
    ShowHint := desc.FQBF.ShowHint;
    Result := (ShowModal = mrOK);
    if Result then
        with desc do begin
            if desc.FQBF.Language = qlFilter then begin
                FArg1 := Var2Text(Val1Edit.DisplayValue);
                FArg2 := Var2Text(Val2Edit.DisplayValue);
            end
            else begin
                FArg1 := Val1Edit.Text;
                FArg2 := Val2Edit.Text;
            end;
            FArg3.Assign(Val3LB.Items);
            FOper := OperDesc[Desc.Ind2OperDesc(SelectOperCB.ItemIndex)].token;
            FNOT := not NOTChb.Checked;
        end;
end;

procedure TNCQBFConditionDlg.AddValBtnClick(Sender: TObject);
var
    Value: String;
begin
    Value := '';
    if InputQuery(LoadStr(SEnterValue), LoadStr(SValue), Value) then
        Val3LB.Items.Add(Value);
end;

procedure TNCQBFConditionDlg.DelValBtnClick(Sender: TObject);
begin
    if YesNoDlg([LoadStr(SDeleteValue)], -1) then
        Val3LB.Items.Delete(Val3LB.ItemIndex);
end;

procedure TNCQBFConditionDlg.EditValBtnClick(Sender: TObject);
var
    Value: String;
begin
    Value := Val3LB.Items[Val3LB.ItemIndex];
    if InputQuery(LoadStr(SEnterValue), LoadStr(SValue), Value) then
        Val3LB.Items[Val3LB.ItemIndex] := Value;
end;

procedure TNCQBFConditionDlg.Val3LBKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    if (Key = VK_F2) or (Key = VK_DOWN) and (ssAlt in Shift) then begin
        EditValBtnClick(nil);
        Key := 0;
    end
    else if (Key = VK_DELETE) then begin
        DelValBtnClick(nil);
        Key := 0;
    end
    else if (Key = VK_INSERT) then begin
        AddValBtnClick(nil);
        Key := 0;
    end;
end;

procedure TNCQBFConditionDlg.NOTChbClick(Sender: TObject);
begin
    NOTChb.Caption := NOTChbCapts[not NOTChb.Checked];
end;

procedure TNCQBFConditionDlg.FormShow(Sender: TObject);
begin
    if Val1Edit.Showing then
        Val1Edit.SetFocus
    else if Val2Edit.Showing then
        Val2Edit.SetFocus
    else if Val3LB.Showing then
        Val3LB.SetFocus
    else
        SelectOperCB.SetFocus;
end;

initialization
    OperDescLoaded := False;
end.
