{***********************************************}
{File:      NCQBFEdt.PAS                        }
{Revision:  2.04 / 04.07.1999                   }
{Comment:   QBF end-user interface dialog       }
{Copyright: (c) 1997-1999, Nica-Com Ltd.        }
{E-Mail:    diman@ncom.nnov.su                  }
{***********************************************}

unit NCQBFEdt;

interface

uses Windows, Classes, Graphics, Forms, Controls, Buttons, StdCtrls, DB,
     ExtCtrls, ComCtrls, Grids, DBTables, NCQBF, NCMemTbl, NCDblLst;

type
  TNCQBFEditDlg = class(TForm)
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    QueryGrid: TStringGrid;
    DblList: TNCDblListBox;
    HelpBtn: TBitBtn;
    CancelBtn: TBitBtn;
    OKBtn: TBitBtn;
    PrevCondBtn: TBitBtn;
    NextCondBtn: TBitBtn;
    AddCondBtn: TBitBtn;
    DelCondBtn: TBitBtn;
    StatusLbl: TLabel;
    Bevel1: TBevel;
    EditCondBtn: TBitBtn;
    ClearCondBtn: TBitBtn;
    Bevel2: TBevel;
    Bevel3: TBevel;
    SortModeChB: TCheckBox;
    TabSheet3: TTabSheet;
    DetailView: TTreeView;
    Bevel4: TBevel;
    EditDetailBtn: TBitBtn;
    CancelDetailBtn: TBitBtn;
    ApplyDetailBtn: TBitBtn;
    ConnectedChB: TCheckBox;
    Label1: TLabel;
    procedure QueryGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure QueryGridDblClick(Sender: TObject);
    procedure PrevCondBtnClick(Sender: TObject);
    procedure NextCondBtnClick(Sender: TObject);
    procedure AddCondBtnClick(Sender: TObject);
    procedure QueryGridKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure EditCondBtnClick(Sender: TObject);
    procedure ClearCondBtnClick(Sender: TObject);
    procedure DelCondBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure QueryGridMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure DblListDstClick(Sender: TObject);
    procedure SortModeChBClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
  private
    { Private declarations }
    UsedFld: Integer;
    QBFTable: TNCMemTable;
    CondNum: Integer;
    FBtnDown: Boolean;

    procedure CommitPredicate;
    procedure LoadPredicate;
    procedure SetupCondLbl;
    procedure GoToCond(dir: Integer);
  public
    { Public declarations }
    function Execute(ds: TDataSet; qt: TNCMemTable; qbf: TNCQBF): Boolean;
  end;

var
  NCQBFEditDlg: TNCQBFEditDlg;

implementation

Uses SysUtils, Dialogs, NCQBFExp, NCStrs, NCUtil, NCDBUtil, NCUIUtil;

{$R *.DFM}

const
    sm2int: array[Boolean] of Integer = (1, -1);

{ ---------------------------------------------------------------------------- }
{ ---------------------------------------------------------------------------- }

procedure TNCQBFEditDlg.QueryGridDrawCell(Sender: TObject; Col,
  Row: Longint; Rect: TRect; State: TGridDrawState);
var
    i: Integer;

    procedure DrawCellText;
    var
        s: String;
    begin
        s := QueryGrid.Cells[Col, Row];
        ExtTextOut(QueryGrid.Canvas.Handle, Rect.Left + 2, Rect.Top + 2,
            ETO_CLIPPED or ETO_OPAQUE, @Rect, PChar(s), Length(s), nil);
    end;

begin
    with QueryGrid.Canvas do begin
        if Row = 0 then begin
            Font.Style := Font.Style + [fsBold];
            DrawCellText;
        end
        else if gdSelected in State then begin
            Rect.Left := Rect.Right - CurrentButtonWidth;
            i := DFCS_SCROLLCOMBOBOX;
            if FBtnDown then
                i := i or DFCS_FLAT or DFCS_PUSHED;
            DrawFrameControl(Handle, Rect, DFC_SCROLL, i);
        end
    end;
end;

procedure TNCQBFEditDlg.QueryGridMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
    R: TRect;
begin
    with QueryGrid do
        if (Col = 1) and (Row >= 1) then begin
            R := CellRect(Col, Row);
            R.Left := R.Right - CurrentButtonWidth;
            if PtInRect(R, Point(X, Y)) then
                QueryGridDblClick(Sender);
        end;
end;

procedure TNCQBFEditDlg.QueryGridDblClick(Sender: TObject);
var
    desc: TNCQBFCondDesc;
begin
    with QueryGrid do
        if Row > 0 then
        try
            FBtnDown := True;
            desc := (Objects[0, Row] as TNCQBFCondDesc);
            if desc.Edit then
                Cells[1, Row] := desc.Description;
        finally
            FBtnDown := False;
        end;
end;

procedure TNCQBFEditDlg.QueryGridKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    if (Key = VK_F2) or (Key = VK_DOWN) and (ssAlt in Shift) then begin
        QueryGridDblClick(Sender);
        Key := 0;
    end
    else if (Key = VK_DELETE) then begin
        ClearCondBtnClick(Sender);
        Key := 0;
    end
    else if (Key = VK_LEFT) and (ssCtrl in Shift) then begin
        PrevCondBtnClick(Self);
        Key := 0;
    end
    else if (Key = VK_RIGHT) and (ssCtrl in Shift) then begin
        NextCondBtnClick(Self);
        Key := 0;
    end
    else if (Key = Ord('A')) and (ssCtrl in Shift) then begin
        AddCondBtnClick(Self);
        Key := 0;
    end
    else if (Key = Ord('D')) and (ssCtrl in Shift) then begin
        DelCondBtnClick(Self);
        Key := 0;
    end;
end;

procedure TNCQBFEditDlg.DblListDstClick(Sender: TObject);
begin
    with DblList do
        if (DstItemIndex >= 0) and (DstItemIndex < DstItems.Count) then
            SortModeChB.Checked := (DstItems.Objects[DstItemIndex] as
                TNCQBFCondDesc).Order < 0
        else
            SortModeChB.State := cbGrayed;
end;

procedure TNCQBFEditDlg.SortModeChBClick(Sender: TObject);
begin
    with DblList do
        if (DstItemIndex >= 0) and (DstItemIndex < DstItems.Count) then
            (DstItems.Objects[DstItemIndex] as TNCQBFCondDesc).Order :=
                sm2int[SortModeChB.Checked];
end;

procedure TNCQBFEditDlg.CommitPredicate;
var
    i: Integer;
    desc: TNCQBFCondDesc;
begin
    QBFTable.Edit;
    for i := 0 to usedFld - 1 do begin
        desc := (QueryGrid.Objects[0, i + 1] as TNCQBFCondDesc);
        QBFTable.FieldByName(desc.field.FieldName).AsString := desc.Condition;
    end;
    QBFTable.Post;
end;

procedure TNCQBFEditDlg.LoadPredicate;
var
    i: Integer;
    desc: TNCQBFCondDesc;
begin
    for i := 0 to usedFld - 1 do begin
        desc := (QueryGrid.Objects[0, i + 1] as TNCQBFCondDesc);
        desc.Condition := QBFTable.FieldByName(desc.field.FieldName).AsString;
        QueryGrid.Cells[1, i + 1] := desc.Description;
    end;
end;

procedure TNCQBFEditDlg.GoToCond(dir: Integer);
begin
    CommitPredicate;
    QBFTable.MoveBy(dir);
    Inc(CondNum, dir);
    SetupCondLbl;
    LoadPredicate;
end;

procedure TNCQBFEditDlg.PrevCondBtnClick(Sender: TObject);
begin
    GoToCond(-1);
end;

procedure TNCQBFEditDlg.NextCondBtnClick(Sender: TObject);
begin
    GoToCond(1);
end;

procedure TNCQBFEditDlg.AddCondBtnClick(Sender: TObject);
begin
    CommitPredicate;
    QBFTable.Append;
    QBFTable.Post;
    CondNum := QBFTable.RecordCount;
    SetupCondLbl;
    LoadPredicate;
end;

procedure TNCQBFEditDlg.DelCondBtnClick(Sender: TObject);
begin
    CommitPredicate;
    try
        QBFTable.Delete;
    except
        SetupCondLbl;
    end;
end;

procedure TNCQBFEditDlg.EditCondBtnClick(Sender: TObject);
begin
    QueryGridDblClick(Sender);
end;

procedure TNCQBFEditDlg.ClearCondBtnClick(Sender: TObject);
begin
    with QueryGrid do
        if Row > 0 then
            with (Objects[0, Row] as TNCQBFCondDesc) do begin
                Clear;
                Cells[1, Row] := Description;
            end;
end;

procedure TNCQBFEditDlg.SetupCondLbl;
begin
    PrevCondBtn.Enabled := not QBFTable.BOF;
    NextCondBtn.Enabled := not QBFTable.EOF;
    if QBFTable.BOF then
        CondNum := 1;
    if QBFTable.EOF then
        CondNum := QBFTable.RecordCount;
    StatusLbl.Caption := IntToStr(CondNum) + ' : ' + IntToStr(QBFTable.RecordCount) +
        ' ' + QBFTable.FieldByName(MTDelMark).AsString;
    if QBFTable.RecExist then
        QueryGrid.Color := clWindow
    else
        QueryGrid.Color := clGray;
end;

procedure TNCQBFEditDlg.FormCreate(Sender: TObject);
begin
    IUStr2Font(LoadStr(SDefaultFont), Font);
    DULoadNavBtnGlyph(nil, PrevCondBtn, NextCondBtn, nil);
    with QueryGrid do begin
        Canvas.Font := Font;
        Cells[0, 0] := LoadStr(SFldCaption);
        Cells[1, 0] := LoadStr(SCondCaption);
    end;
end;

procedure TNCQBFEditDlg.FormShow(Sender: TObject);
begin
    SetupCondLbl;
end;

procedure TNCQBFEditDlg.FormResize(Sender: TObject);
var
    dY, dX: Integer;
begin
    LockWindowUpdate(Handle);
    try
        dY := (ClientHeight - (342 - 301)) - PageControl1.Height;
        dX := (ClientWidth - (430 - 430)) - PageControl1.Width;
        PageControl1.Height := PageControl1.Height + dY;
        PageControl1.Width := PageControl1.Width + dX;
        QueryGrid.Height := QueryGrid.Height + dY;
        QueryGrid.Width := QueryGrid.Width + dX;
        QueryGrid.ColWidths[0] := QueryGrid.ColWidths[0] + dX;
        Bevel2.Width := Bevel2.Width + dX;
        DblList.Height := DblList.Height + dY;
        DblList.Width := DblList.Width + dX;
        Bevel3.Width := Bevel3.Width + dX;
        OKBtn.Top := OKBtn.Top + dY;
        CancelBtn.Top := CancelBtn.Top + dY;
        HelpBtn.Top := HelpBtn.Top + dY;
    finally
        LockWindowUpdate(0);
    end;
end;

function TNCQBFEditDlg.Execute(ds: TDataSet; qt: TNCMemTable; qbf: TNCQBF): Boolean;
var
    j, i: Integer;
    h, maxLabelH, w, maxLabelW: Integer;
    desc: TNCQBFCondDesc;
    unionEnabledOper, enabledOper: TQBFTokens;
    dispLbl: String;
begin
    QBFTable := qt;
    QBFTable.Post;
    try
        with QueryGrid do begin
            HandleNeeded;
            RowCount := ds.FieldCount + 1;
            maxLabelW := 0;
            maxLabelH := 0;
            usedFld := 0;
            unionEnabledOper := [];
            for i := 0 to ds.FieldCount - 1 do begin
                dispLbl := StrTran(ds.Fields[i].DisplayLabel, ';', ' ');
                enabledOper := qbf.CanQueryField(ds.Fields[i]);
                if enabledOper = [] then
                    Continue;
                unionEnabledOper := unionEnabledOper + enabledOper;
                desc := TNCQBFCondDesc.Create(qbf, ds.Fields[i]);
                desc.Condition := QBFTable.FieldByName(ds.Fields[i].FieldName).AsString;
                if enabledOper - [tkSORTASC, tkSORTDESC] <> [] then begin
                    Cells[0, usedFld + 1] := dispLbl;
                    Objects[0, usedFld + 1] := desc;
                    Cells[1, usedFld + 1] := desc.Description;
                    w := Canvas.TextWidth(dispLbl);
                    h := Canvas.TextHeight(dispLbl);
                    if w > maxLabelW then
                        maxLabelW := w;
                    if h > maxLabelH then
                        maxLabelH := h;
                    Inc(usedFld);
                end;
                if [tkSORTASC, tkSORTDESC] * enabledOper <> [] then begin
                    if desc.Order <> 0 then begin
                        j := 0;
                        while (j <= DblList.DstItems.Count - 1) and
                              (abs((DblList.DstItems.Objects[j] as TNCQBFCondDesc).Order) <= abs(desc.Order)) do
                            Inc(j);
                        if j > DblList.DstItems.Count - 1 then
                            j := DblList.DstItems.Count - 1;
                        DblList.DstItems.InsertObject(j, dispLbl, desc);
                    end
                    else
                        DblList.SrcItems.AddObject(dispLbl, desc);
                end;
            end;
            if usedFld > 0 then begin
                RowCount := usedFld + 1;
                if maxLabelW > ClientWidth div 2 then
                    maxLabelW := ClientWidth div 2;
                ColWidths[0] := maxLabelW + 4;
                ColWidths[1] := ClientWidth - maxLabelW - 6;
                DefaultRowHeight := maxLabelH + 4;
                TabSheet1.TabVisible := True;
                PageControl1.ActivePage := TabSheet1;
            end
            else begin
                TabSheet1.TabVisible := False;
                PageControl1.ActivePage := TabSheet2;
            end;
        end;
        TabSheet2.TabVisible := (unionEnabledOper * [tkSORTASC, tkSORTDESC]) <> [];
        SortModeChB.Checked := (unionEnabledOper * [tkSORTASC, tkSORTDESC]) = [tkSORTDESC];
        SortModeChB.Enabled := (unionEnabledOper * [tkSORTASC, tkSORTDESC]) = [tkSORTASC, tkSORTDESC];
        Result := False;
        ShowHint := qbf.ShowHint;
        QBFTable.CheckBrowseMode;
        while not QBFTable.BOF do
            QBFTable.Prior;
        CondNum := 1;
        if ShowModal = mrOK then begin
            for i := 0 to DblList.DstItems.Count - 1 do
                (DblList.DstItems.Objects[i] as TNCQBFCondDesc).Order :=
                    sm2int[(DblList.DstItems.Objects[i] as TNCQBFCondDesc).Order < 0] * (i + 1);
            for i := 0 to DblList.SrcItems.Count - 1 do
                (DblList.SrcItems.Objects[i] as TNCQBFCondDesc).Order := 0;
            CommitPredicate;
            Result := True;
        end;
    finally
        for i := 0 to usedFld - 1 do begin
            QueryGrid.Objects[0, i + 1].Free;
            QueryGrid.Cells[0, i + 1] := '';
            QueryGrid.Cells[1, i + 1] := '';
        end;
        DblList.SrcItems.Clear;
        DblList.DstItems.Clear;
    end;
end;

end.
