{*******************************************************}
{File:      NCOciBreakDlg.pas                           }
{Revision:  0.02.00 / 28.08.2000                        }
{Comment:   NC OCI8 VCL: Oracle8 break query dialog     }
{Copyright: (c) 1999-2000, Dmitry Arefiev               }
{Author:    Dmitry Arefiev, dmitrya@inthink.com         }
{*******************************************************}
{$I NCOciDef.inc}

unit NCOciBreakDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, NCOci,
  NCOciDB, StdCtrls, Buttons, ExtCtrls;

type
  TNCOciBreakFrm = class(TForm)
    BitBtn1: TBitBtn;
    Timer: TTimer;
    procedure Button1Click(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure TimerTimer(Sender: TObject);
  private
    { Private declarations }
    FDatabase: TOCIDatabase;
    FRequestShow: Boolean;
    procedure DoRequest(AShow: Boolean; ADatabase: TOCIDatabase);
  public
    { Public declarations }
    class procedure Execute(ADatabase: TOCIDatabase);
    class procedure Cancel(ADatabase: TOCIDatabase);
    class procedure ProcessMessages;
  end;

var
  NCOciBreakFrm: TNCOciBreakFrm;

implementation

{$R *.DFM}

class procedure TNCOciBreakFrm.Execute(ADatabase: TOCIDatabase);
begin
    if NCOciBreakFrm = nil then
        NCOciBreakFrm := TNCOciBreakFrm.Create(nil);
    NCOciBreakFrm.DoRequest(True, ADatabase);
end;

class procedure TNCOciBreakFrm.Cancel(ADatabase: TOCIDatabase);
begin
    if NCOciBreakFrm <> nil then
        NCOciBreakFrm.DoRequest(False, ADatabase);
end;

class procedure TNCOciBreakFrm.ProcessMessages;
var
    Msg: TMsg;

    function GetMouseMessageForm(const Msg: TMsg): TCustomForm;
    var
        P: TPoint;
        ctrl: TControl;
    begin
        P := Point(LoWord(Msg.lParam), HiWord(Msg.lParam));
        ctrl := FindDragTarget(P, True);
        if Ctrl <> nil then
            Result := GetParentForm(ctrl)
        else
            Result := nil;
    end;

begin
    while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do begin
        if not (
            (Msg.message >= WM_KEYDOWN) and (Msg.message <= WM_DEADCHAR) and
                (Screen.ActiveForm <> NCOciBreakFrm) or
            (Msg.message >= WM_MOUSEFIRST) and (Msg.message <= WM_MOUSELAST) and
                (GetMouseMessageForm(Msg) <> NCOciBreakFrm) or
            (Msg.message >= WM_SYSKEYDOWN) and (Msg.message <= WM_SYSDEADCHAR)
           ) then begin
            TranslateMessage(Msg);
            DispatchMessage(Msg);
        end;
    end;
end;

procedure TNCOciBreakFrm.DoRequest(AShow: Boolean; ADatabase: TOCIDatabase);
begin
    FDatabase := ADatabase;
    if (FDatabase = nil) or (csDestroying in FDatabase.ComponentState) then begin
        Free;
        NCOciBreakFrm := nil;
    end
    else if (AShow <> FRequestShow) and (FDatabase <> nil) and
            (FDatabase.Connected or not AShow) then begin
        FRequestShow := AShow;
        Timer.Interval := WDelayBeforeSQLCrs;
        Timer.Enabled := True;
    end;
end;

procedure TNCOciBreakFrm.TimerTimer(Sender: TObject);
begin
    if FRequestShow <> Visible then begin
        if FRequestShow then begin
            if GetCapture <> 0 then
                SendMessage(GetCapture, WM_CANCELMODE, 0, 0);
            Show;
        end
        else
            Hide;
        Timer.Enabled := False;
        TNCOciBreakFrm.ProcessMessages;
    end;
end;

procedure TNCOciBreakFrm.Button1Click(Sender: TObject);
begin
    FDatabase.Break;
end;

procedure TNCOciBreakFrm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
    if Key = VK_ESCAPE then begin
        Button1Click(nil);
        Key := 0;
    end;
end;

initialization
    NCOciBreakFrm := nil;

finalization
    NCOciBreakFrm.Free;
    NCOciBreakFrm := nil;
    
end.
