{*********************************************************}
{File:      NCOciMove.PAS                                 }
{Revision:  0.01.00 / 01.09.2000                          }
{Comment:   NC OCI8 VCL: Oracle8 batch moving services    }
{Copyright: (c) 1999-2000, Dmitry Arefiev                 }
{Author:    Dmitry Arefiev, dmitrya@inthink.com           }
{*********************************************************}
{$I NCOciDef.inc}

unit NCOciMove;

interface

Uses SysUtils, Classes, NCOci, NCOciWrapper, NCOciDB;

type
    TOCIDPColumn = class;
    TOCIDPColumns = class;
    TOCIDirectLoader = class;

    TOCIDPColumn = class(TCollectionItem)
    private
        FName: String;
        FDataType: TOCIDirecPathDataType;
        FDataSize: Integer;
        FScale: Integer;
        FPrecision: Integer;
        FDateFormat: String;
        procedure SetDataType(const Value: TOCIDirecPathDataType);
        function GetDateTimes(AIndex: Integer): TDateTime;
        function GetFloats(AIndex: Integer): Double;
        function GetIntegers(AIndex: Integer): Integer;
        function GetIsNulls(AIndex: Integer): Boolean;
        function GetStrings(AIndex: Integer): String;
        function GetUIntegers(AIndex: Integer): Cardinal;
        procedure SetDateTimes(AIndex: Integer; const Value: TDateTime);
        procedure SetFloats(AIndex: Integer; const Value: Double);
        procedure SetIntegers(AIndex: Integer; const Value: Integer);
        procedure SetIsNulls(AIndex: Integer; const Value: Boolean);
        procedure SetStrings(AIndex: Integer; const Value: String);
        procedure SetUIntegers(AIndex: Integer; const Value: Cardinal);
        function GetDirectLoader: TOCIDirectLoader;
        procedure CheckAccess(ATypes: TOCIDirecPathDataTypes);
    protected
        function GetDisplayName: string; override;
    public
        constructor Create(Collection: TCollection); override;
        procedure Assign(AValue: TPersistent); override;
        procedure SetRaws(AIndex: Integer; AData: Pointer; ASize: ub4);
        procedure GetRaws(AIndex: Integer; AData: Pointer; var ALen: ub4);
        property AsStrings[AIndex: Integer]: String read GetStrings write SetStrings;
        property AsDateTimes[AIndex: Integer]: TDateTime read GetDateTimes write SetDateTimes;
        property AsIntegers[AIndex: Integer]: Integer read GetIntegers write SetIntegers;
        property AsUIntegers[AIndex: Integer]: Cardinal read GetUIntegers write SetUIntegers;
        property AsFloats[AIndex: Integer]: Double read GetFloats write SetFloats;
        property IsNulls[AIndex: Integer]: Boolean read GetIsNulls write SetIsNulls;
        property DirectLoader: TOCIDirectLoader read GetDirectLoader;
    published
        property Name: String read FName write FName;
        property DataType: TOCIDirecPathDataType read FDataType write SetDataType default dtString;
        property DataSize: Integer read FDataSize write FDataSize default 255;
        property Scale: Integer read FScale write FScale default 0;
        property Precision: Integer read FPrecision write FPrecision default 0;
        property DateFormat: String read FDateFormat write FDateFormat;
    end;

    TOCIDPColumns = class(TCollection)
    private
        FOwner: TPersistent;
        function GetItem(Index: Integer): TOCIDPColumn;
        procedure SetItem(Index: Integer; const Value: TOCIDPColumn);
    protected
        function GetOwner: TPersistent; override;
    public
        constructor Create(AOwner: TPersistent); {$IFDEF OCI_D4} overload; {$ENDIF}
        function Add: TOCIDPColumn;
        function FindColumn(const AName: String): TOCIDPColumn;
        function ColumnByName(const AName: String): TOCIDPColumn;
        property Items[Index: Integer]: TOCIDPColumn read GetItem write SetItem; default;
    end;

    TOCIDirectLoader = class(TOCICustomWrappedComponent)
    private
        FTableName: String;
        FDateFormat: String;
        FLogging: Boolean;
        FParallel: Boolean;
        FBufferSize: Integer;
        FColumns: TOCIDPColumns;
        FDatabaseName: String;
        FDatabase: TOCIDatabase;
        FDirectPath: TOCIDirectPath;
        FColArray: TOCIDirectPathColArray;
        FStream: TOCIDirectPathStream;
        FOPartitionName: String;
        FOTableName: String;
        FOSubPartitionName: String;
        FRowsCount: Integer;
        FPrepared: Boolean;
        procedure SetColumns(const Value: TOCIDPColumns);
        procedure SetDatabaseName(const Value: String);
        procedure SetTableName(const Value: String);
        procedure SplitName(const AName: String; var AOwnerName,
            ATblName, APartName, ASubPartName: String);
        procedure ObjChanged(AFullName: Boolean);
        procedure SetOPartitionName(const Value: String);
        procedure SetOSubPartitionName(const Value: String);
        procedure SetOTableName(const Value: String);
        procedure CheckInactive;
    protected
        function GetEntry(ARow, ACol: Integer; AData: pUb1; var ASize: ub4): Boolean;
        procedure SetEntry(ARow, ACol: Integer; AData: pUb1; ASize: ub4);
    public
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        function FindColumn(const AName: String): TOCIDPColumn;
        function ColumnByName(const AName: String): TOCIDPColumn;
        procedure Describe;
        procedure Prepare;
        procedure Execute;
        property Database: TOCIDatabase read FDatabase;
        property HDirectPath: TOCIDirectPath read FDirectPath;
        property HColArray: TOCIDirectPathColArray read FColArray;
        property HStream: TOCIDirectPathStream read FStream;
        property RowsCount: Integer read FRowsCount;
        property Prepared: Boolean read FPrepared;
    published
        property DatabaseName: String read FDatabaseName write SetDatabaseName;
        property TableName: String read FTableName write SetTableName stored False;
        property OTableName: String read FOTableName write SetOTableName;
        property OPartitionName: String read FOPartitionName write SetOPartitionName;
        property OSubPartitionName: String read FOSubPartitionName write SetOSubPartitionName;
        property Columns: TOCIDPColumns read FColumns write SetColumns;
        property DateFormat: String read FDateFormat write FDateFormat;
        property Logging: Boolean read FLogging write FLogging default True;
        property Parallel: Boolean read FParallel write FParallel default False;
        property BufferSize: Integer read FBufferSize write FBufferSize default 65536;
    end;

implementation

Uses NCOciMsg;

const
    varDT2dirDT: array[TOCIVarDataType] of TOCIDirecPathDataType = (
        dtUnknown, dtInteger, dtInteger, dtFloat, dtFloat, dtString,
        dtString, dtString, dtRaw, dtDateTime, dtRaw, dtRaw, dtUnknown,
        dtUnknown, dtRaw, dtRaw, dtRaw, dtRaw, dtUnknown);

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// TOCIDPColumn

constructor TOCIDPColumn.Create(Collection: TCollection);
begin
    inherited Create(Collection);
    DataType := dtString;
end;

procedure TOCIDPColumn.Assign(AValue: TPersistent);
begin
    if AValue is TOCIDPColumn then
        try
            if Collection <> nil then
                Collection.BeginUpdate;
            Name := TOCIDPColumn(AValue).Name;
            DataType := TOCIDPColumn(AValue).DataType;
            DataSize := TOCIDPColumn(AValue).DataSize;
            Scale := TOCIDPColumn(AValue).Scale;
            Precision := TOCIDPColumn(AValue).Precision;
            DateFormat := TOCIDPColumn(AValue).DateFormat;
        finally
            if Collection <> nil then
                Collection.EndUpdate;
        end
    else
        inherited Assign(AValue);
end;

function TOCIDPColumn.GetDisplayName: string;
begin
    if Name <> '' then
        Result := Name
    else
        Result := inherited GetDisplayName;
end;

procedure TOCIDPColumn.SetDataType(const Value: TOCIDirecPathDataType);
begin
    if FDataType <> Value then begin
        FDataType := Value;
        DataSize := nc2ociDPValueSize[Value];
        if Value in [dtString, dtDateTime, dtRaw, dtUnknown] then begin
            Precision := 0;
            Scale := 0;
        end
        else if Value in [dtInteger, dtUInteger] then
            Scale := 0;
    end;
end;

function TOCIDPColumn.GetDirectLoader: TOCIDirectLoader;
begin
    Result := TOCIDPColumns(Collection).FOwner as TOCIDirectLoader;
end;

function TOCIDPColumn.GetStrings(AIndex: Integer): String;
var
    len: ub4;
begin
    CheckAccess([dtString, dtRaw]);
    SetLength(Result, DataSize);
    Result := '';
    if DataSize > 0 then
        if DirectLoader.GetEntry(AIndex, Index, @Result[1], len) then
            SetLength(Result, len);
end;

procedure TOCIDPColumn.SetStrings(AIndex: Integer; const Value: String);
begin
    CheckAccess([dtString, dtRaw]);
    if Value = '' then
        DirectLoader.SetEntry(AIndex, Index, nil, 0)
    else
        DirectLoader.SetEntry(AIndex, Index, @Value[1], Length(Value));
end;

function TOCIDPColumn.GetDateTimes(AIndex: Integer): TDateTime;
var
    dt: TOraDate;
    len: ub4;
begin
    CheckAccess([dtDateTime]);
    if DirectLoader.GetEntry(AIndex, Index, @dt, len) then
        OraDate2Delphi(@dt, Result)
    else
        Result := 0.0;
end;

procedure TOCIDPColumn.SetDateTimes(AIndex: Integer; const Value: TDateTime);
var
    dt: TOraDate;
begin
    CheckAccess([dtDateTime]);
    Delphi2OraDate(@dt, Value);
    DirectLoader.SetEntry(AIndex, Index, @dt, SizeOf(TOraDate));
end;

function TOCIDPColumn.GetFloats(AIndex: Integer): Double;
var
    len: ub4;
begin
    CheckAccess([dtFloat]);
    if not DirectLoader.GetEntry(AIndex, Index, @Result, len) then
        Result := 0.0;
end;

procedure TOCIDPColumn.SetFloats(AIndex: Integer; const Value: Double);
begin
    CheckAccess([dtFloat]);
    DirectLoader.SetEntry(AIndex, Index, @Value, SizeOf(Double));
end;

function TOCIDPColumn.GetIntegers(AIndex: Integer): Integer;
var
    len: ub4;
begin
    CheckAccess([dtInteger, dtUInteger]);
    if not DirectLoader.GetEntry(AIndex, Index, @Result, len) then
        Result := 0;
end;

procedure TOCIDPColumn.SetIntegers(AIndex: Integer; const Value: Integer);
begin
    CheckAccess([dtInteger, dtUInteger]);
    DirectLoader.SetEntry(AIndex, Index, @Value, SizeOf(Integer));
end;

function TOCIDPColumn.GetUIntegers(AIndex: Integer): Cardinal;
var
    len: ub4;
begin
    CheckAccess([dtInteger, dtUInteger]);
    if not DirectLoader.GetEntry(AIndex, Index, @Result, len) then
        Result := 0;
end;

procedure TOCIDPColumn.SetUIntegers(AIndex: Integer; const Value: Cardinal);
begin
    CheckAccess([dtInteger, dtUInteger]);
    DirectLoader.SetEntry(AIndex, Index, @Value, SizeOf(Cardinal));
end;

procedure TOCIDPColumn.GetRaws(AIndex: Integer; AData: Pointer; var ALen: ub4);
begin
    CheckAccess([dtString, dtRaw]);
    DirectLoader.GetEntry(AIndex, Index, AData, ALen);
end;

procedure TOCIDPColumn.SetRaws(AIndex: Integer; AData: Pointer; ASize: ub4);
begin
    CheckAccess([dtString, dtRaw]);
    DirectLoader.SetEntry(AIndex, Index, AData, ASize);
end;

function TOCIDPColumn.GetIsNulls(AIndex: Integer): Boolean;
var
    len: ub4;
begin
    CheckAccess([dtString, dtDateTime, dtInteger, dtUInteger, dtFloat, dtRaw]);
    Result := not DirectLoader.GetEntry(AIndex, Index, nil, len);
end;

procedure TOCIDPColumn.SetIsNulls(AIndex: Integer; const Value: Boolean);
begin
    CheckAccess([dtString, dtDateTime, dtInteger, dtUInteger, dtFloat, dtRaw]);
    if Value then
        DirectLoader.SetEntry(AIndex, Index, nil, 0);
end;

procedure TOCIDPColumn.CheckAccess(ATypes: TOCIDirecPathDataTypes);
begin
    if not (DataType in ATypes) or (Collection = nil) or
       (DirectLoader = nil) or not DirectLoader.Prepared then
        OCIDBError(msgOCIDPColumnAccessError, nil);
end;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// TOCIDPColumns

constructor TOCIDPColumns.Create(AOwner: TPersistent);
begin
    inherited Create(TOCIDPColumn);
    FOwner := AOwner;
end;

function TOCIDPColumns.Add: TOCIDPColumn;
begin
    Result := TOCIDPColumn(inherited Add);
end;

function TOCIDPColumns.GetItem(Index: Integer): TOCIDPColumn;
begin
    Result := TOCIDPColumn(inherited GetItem(Index));
end;

procedure TOCIDPColumns.SetItem(Index: Integer; const Value: TOCIDPColumn);
begin
    inherited SetItem(Index, Value);
end;

function TOCIDPColumns.GetOwner: TPersistent;
begin
    Result := FOwner;
end;

function TOCIDPColumns.FindColumn(const AName: String): TOCIDPColumn;
var
    i: Integer;
begin
    Result := nil;
    for i := 0 to Count - 1 do
        if AnsiCompareText(Items[i].Name, AName) = 0 then begin
            Result := Items[i];
            Break;
        end;
end;

function TOCIDPColumns.ColumnByName(const AName: String): TOCIDPColumn;
begin
    Result := FindColumn(AName);
    if Result = nil then
        OCIDBError(msgOCIDPColumnNotFound, nil);
end;

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------
// TOCIDirectLoader

constructor TOCIDirectLoader.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    FColumns := TOCIDPColumns.Create(Self);
end;

destructor TOCIDirectLoader.Destroy;
begin
    FColumns.Free;
    inherited Destroy;
end;

procedure TOCIDirectLoader.CheckInactive;
begin
    if Prepared then
        OCIDBError(msgOCIDPMustBeUnprepared, nil);
end;

procedure TOCIDirectLoader.SetColumns(const Value: TOCIDPColumns);
begin
    CheckInactive;
    FColumns.Assign(Value);
end;

function TOCIDirectLoader.FindColumn(const AName: String): TOCIDPColumn;
begin
    Result := FColumns.FindColumn(AName);
end;

function TOCIDirectLoader.ColumnByName(const AName: String): TOCIDPColumn;
begin
    Result := FColumns.ColumnByName(AName);
end;

procedure TOCIDirectLoader.SetDatabaseName(const Value: String);
begin
    CheckInactive;
    if FDatabaseName <> Value then begin
        FDatabaseName := Value;
        if csDesigning in ComponentState then
            Describe;
    end;
end;

procedure TOCIDirectLoader.SetTableName(const Value: String);
var
    oraOwner: String;
begin
    CheckInactive;
    if FTableName <> Value then begin
        FTableName := Value;
        SplitName(Value, oraOwner, FOTableName, FOPartitionName, FOSubPartitionName);
        if oraOwner <> '' then
            FOTableName := oraOwner + '.' + FOTableName;
        ObjChanged(True);
    end;
end;

procedure TOCIDirectLoader.SetOPartitionName(const Value: String);
begin
    CheckInactive;
    if FOPartitionName <> Value then begin
        FOPartitionName := Value;
        ObjChanged(False);
    end;
end;

procedure TOCIDirectLoader.SetOSubPartitionName(const Value: String);
begin
    CheckInactive;
    if FOSubPartitionName <> Value then begin
        FOSubPartitionName := Value;
        ObjChanged(False);
    end;
end;

procedure TOCIDirectLoader.SetOTableName(const Value: String);
begin
    CheckInactive;
    if FOTableName <> Value then begin
        FOTableName := Value;
        ObjChanged(False);
    end;
end;

procedure TOCIDirectLoader.ObjChanged(AFullName: Boolean);
begin
    if not AFullName then begin
        FTableName := FOTableName;
        if FOPartitionName <> '' then
            FTableName := FTableName + ' PARTITION(' + FOPartitionName + ')'
        else if FOSubPartitionName <> '' then
            FTableName := FTableName + ' SUBPARTITION(' + FOSubPartitionName + ')';
    end;
    if csDesigning in ComponentState then
        Describe;
end;

// [<schema>.] <table> [ @<dblink> | PARTITION (<pName>) | SUBPARTITION (<subpName>) ]
procedure TOCIDirectLoader.SplitName(const AName: String; var AOwnerName,
    ATblName, APartName, ASubPartName: String);
var
    InName: Boolean;
    pCh, pStName: PChar;
    i, n: Integer;
    obj: array[0..8] of String;
begin
    InName := False;
    pStName := PChar(AName);
    pCh := pStName - 1;
    i := 0;
    repeat
        Inc(pCh);
        case pCh^ of
        '"':
            InName := not InName;
        '.', '@', ' ', #7, #13, #10, ')', '(':
            if not InName then begin
                if i > 7 then
                    OCIDBError(msgOCIDPTabNameError, Self);
                while pCh^ in [' ', #7, #13, #10] do
                    Inc(pCh);
                if not (pCh^ in ['.', '@', ')', '(']) then
                    Dec(pCh^);
                Obj[i] := pCh^;
                Inc(i);
                SetString(Obj[i], pStName, pCh - pStName);
                Obj[i] := Trim(Obj[i]);
                Inc(i);
                pStName := pCh + 1;
            end;
        end;
    until (pCh^ = #0);
    n := i;
    i := 0;
    if (n > 1) and (obj[i + 1] = '.') then begin
        AOwnerName := obj[i];
        Inc(i, 2);
    end;
    ATblName := obj[i];
    Inc(i);
    if (n > (i + 1)) and (obj[i + 1] = '@') then
        OCIDBError(msgOCIDPNotRemote, Self);
    if (n > (i + 4)) and (obj[i + 2] = '(') and (obj[i + 4] = ')') and (
        (CompareText(obj[i + 1], 'PARTITION') = 0) or (CompareText(obj[i + 1], 'SUBPARTITION') = 0)
       ) then begin
        if CompareText(obj[i + 1], 'PARTITION') = 0 then
            APartName := obj[i + 3]
        else
            ASubPartName := obj[i + 3];
        Inc(i, 5);
    end;
    if i < n then
        OCIDBError(msgOCIDPTabNameError, Self);
end;

procedure TOCIDirectLoader.Describe;
var
    i: Integer;
    dt: TOCIVarDataType;
    ds: ub4;
begin
    FColumns.BeginUpdate;
    try
        FColumns.Clear;
        with TOCIQuery.Create(nil) do
        try
            DatabaseName := Self.DatabaseName;
            SQL.Add('select * from ' + TableName + ' where 0 = 1');
            Open;
            for i := 0 to OFieldDefs.Count - 1 do
                with FColumns.Add do begin
                    Name := OFieldDefs[i].Name;
                    dt := OFieldDefs[i].DataType;
                    ds := OFieldDefs[i].DataSize;
                    DataType := varDT2dirDT[dt];
                    if dt = otNumber then
                        DataSize := 40
                    else if (dt in [otLong, otLongRaw, otCLOB, otBLOB, otCFile, otBFile]) and
                            (ds = 0) then
                        DataSize := 4096
                    else
                        DataSize := ds;
                    Scale := OFieldDefs[i].Scale;
                    Precision := OFieldDefs[i].Precision;
                end;
        finally
            Free;
        end;
    finally
        FColumns.EndUpdate;
    end;
end;

procedure TOCIDirectLoader.Prepare;
var
    i: Integer;
    col: TOCIDirectPathColumnParam;
    ownerName, tblName, subName: String;
begin
    FDatabase := TOCIDatabase.DatabaseByName(DatabaseName, Self);
    FDirectPath := TOCIDirectPath.Create(Database.HService);
    SplitName(TableName, ownerName, tblName, subName, subName);
    FDirectPath.SCHEMA_NAME := ownerName;
    FDirectPath.NAME := tblName;
    FDirectPath.SUB_NAME := subName;
    FDirectPath.DATEFORMAT := DateFormat;
    FDirectPath.DIRPATH_MODE := OCI_DIRPATH_LOAD;
    FDirectPath.DIRPATH_NOLOG := Integer(not Logging);
    FDirectPath.DIRPATH_PARALLEL := Integer(Parallel);
    FDirectPath.NUM_COLS := FColumns.Count;
    for i := 0 to FColumns.Count - 1 do begin
        col := FDirectPath.Columns[i];
        try
            col.NAME := FColumns[i].Name;
            col.DataType := FColumns[i].DataType;
            col.DATA_SIZE := FColumns[i].DataSize;
            col.SCALE := FColumns[i].Scale;
            col.PRECISION := FColumns[i].Precision;
            col.DATEFORMAT := FColumns[i].DateFormat;
        finally
            col.Free;
        end;
    end;
    FDirectPath.BUF_SIZE := BufferSize;
    FDirectPath.Prepare;
    FColArray := TOCIDirectPathColArray.Create(FDirectPath);
    FColArray.Reset;
    FRowsCount := FColArray.NUM_ROWS;
    FStream := TOCIDirectPathStream.Create(FDirectPath);
    FStream.Reset;
    FPrepared := True;
end;

function TOCIDirectLoader.GetEntry(ARow, ACol: Integer; AData: pUb1; var ASize: ub4): Boolean;
var
    pData: pUb1;
    flg: ub1;
begin
    Result := False;
    ASize := 0;
    FColArray.EntryGet(ARow, ACol, pData, ASize, flg);
    if flg = OCI_DIRPATH_COL_COMPLETE then begin
        Result := True;
        if AData <> nil then
            Move(pData^, AData^, ASize);
    end;
end;

procedure TOCIDirectLoader.SetEntry(ARow, ACol: Integer; AData: pUb1; ASize: ub4);
var
    flg: ub1;
begin
    if AData = nil then
        flg := OCI_DIRPATH_COL_NULL
    else
        flg := OCI_DIRPATH_COL_COMPLETE;
    FColArray.EntryGet(ARow, ACol, AData, ASize, flg);
end;

procedure TOCIDirectLoader.Load(ARows: Integer);
begin
    repeat
        res := FColArray.ToStream(FStream, ARows, 0);
        FDirectPath.LoadStream(FStream);
        FStream.Reset;
    until res = OCI_SUCCESS;
end;

end.
