{*******************************************************}
{File:      NCOciReg.PAS                                }
{Revision:  0.05.01 / 01.11.2000                        }
{Comment:   NC OCI8 VCL: registration unit              }
{Copyright: (c) 1999-2000, Dmitry Arefiev               }
{Author:    Dmitry Arefiev, dmitrya@inthink.com         }
{*******************************************************}
{$I NCOciDef.inc}

unit NCOciReg;

interface

    procedure Register;

implementation

{$R NCO8.DCR}

Uses Classes, DB, NCOciDB, DsgnIntf, NCOCIMsg, NCSQLMon, NCOciUpdateSQL;

{ --------------------------------------------------------------------------- }
{ --------------------------------------------------------------------------- }

Type
    TOCIListProperty = class(TStringProperty)
        FObjKind: TOCIObjKind;
        FParentObj: String;
        function GetAttributes: TPropertyAttributes; override;
        procedure GetValues(Proc: TGetStrProc); override;
    end;

    TOCIServerNameProperty = class(TOCIListProperty)
        procedure GetValues(Proc: TGetStrProc); override;
    end;

    TOCIDatabaseNameProperty = class(TOCIListProperty)
        procedure GetValues(Proc: TGetStrProc); override;
    end;

    TOCIStoredProcNameProperty = class(TOCIListProperty)
        procedure GetValues(Proc: TGetStrProc); override;
    end;

    TOCIOPackageNameProperty = class(TOCIListProperty)
        procedure GetValues(Proc: TGetStrProc); override;
    end;

    TOCIOProcedureNameProperty = class(TOCIListProperty)
        procedure GetValues(Proc: TGetStrProc); override;
    end;

{$IFDEF OCI_D5}
    TOCILoginCategory = class(TPropertyCategory)
    public
        class function Name: String; override;
        class function Description: String; override;
    end;

    TOCIResourceCategory = class(TPropertyCategory)
    public
        class function Name: String; override;
        class function Description: String; override;
    end;

    TOCITransactCategory = class(TPropertyCategory)
    public
        class function Name: String; override;
        class function Description: String; override;
    end;
{$ENDIF}

    TOCISequenceNameProperty = class(TOCIListProperty)
        procedure GetValues(Proc: TGetStrProc); override;
    end;

{ --------------------------------------------------------------------------- }
{ --------------------------------------------------------------------------- }

function TOCIListProperty.GetAttributes: TPropertyAttributes;
begin
    Result := [paValueList, paSortList, paMultiSelect];
end;

procedure TOCIListProperty.GetValues(Proc: TGetStrProc);
var
    List: TStringList;
    i: Integer;
    db: String;
    comp: TPersistent;
begin
    List := TStringList.Create;
    try
        comp := GetComponent(0);
        db := '';
        if not (FObjKind in [okService, okDatabase]) then
            if comp is TOCIDatabase then
                db := TOCIDatabase(comp).DatabaseName
            else if comp is TOCIDataSet then
                db := TOCIDataSet(comp).DatabaseName
            else if comp is TOCITransactionManager then
                db := TOCITransactionManager(comp).DatabaseName
            else if comp is TOCISequence then
                db := TOCISequence(comp).DatabaseName;
        TOCIDatabase.GetObjectsList(db, List, FParentObj, FObjKind, True);
        for i := 0 to List.Count - 1 do
            Proc(List[i]);
    finally
        List.Free;
    end;
end;

procedure TOCIServerNameProperty.GetValues(Proc: TGetStrProc);
begin
    FObjKind := okService;
    FParentObj := '';
    inherited GetValues(Proc);
end;

procedure TOCIDatabaseNameProperty.GetValues(Proc: TGetStrProc);
begin
    FObjKind := okDatabase;
    FParentObj := '';
    inherited GetValues(Proc);
end;

procedure TOCIStoredProcNameProperty.GetValues(Proc: TGetStrProc);
begin
    FObjKind := okProc;
    FParentObj := '';
    inherited GetValues(Proc);
end;

procedure TOCIOPackageNameProperty.GetValues(Proc: TGetStrProc);
begin
    FObjKind := okPackage;
    FParentObj := '';
    inherited GetValues(Proc);
end;

procedure TOCIOProcedureNameProperty.GetValues(Proc: TGetStrProc);
begin
    FObjKind := okProc;
    FParentObj := (GetComponent(0) as TOCIStoredProc).OPackageName;
    inherited GetValues(Proc);
end;

procedure TOCISequenceNameProperty.GetValues(Proc: TGetStrProc);
begin
    FObjKind := okSeqs;
    FParentObj := (GetComponent(0) as TOCISequence).SequenceName;
    inherited GetValues(Proc);
end;

{ --------------------------------------------------------------------------- }
{ --------------------------------------------------------------------------- }

var
    DataSetEditorClass: TComponentEditorClass;

{$WARNINGS OFF}
procedure CaptureDataSetEditorClass;
var
    ds: TDataSet;
    dsEditor: TComponentEditor;
begin
    ds := nil;
    dsEditor := nil;
    try
        ds := TDataSet.Create(nil);
        dsEditor := GetComponentEditor(ds, nil);
        DataSetEditorClass := TComponentEditorClass(dsEditor.ClassType);
    finally
        dsEditor.Free;
        ds.Free;
    end;
end;
{$WARNINGS ON}

type
    TOCIStatementDataSetEditor = class(TComponentEditor)
    private
        FPrevEditor: TComponentEditor;
    public
        constructor Create(AComponent: TComponent; ADesigner:
            {$IFDEF OCI_D4} IFormDesigner {$ELSE} TFormDesigner {$ENDIF}); override;
        destructor Destroy; override;
        procedure ExecuteVerb(Index: Integer); override;
        function GetVerb(Index: Integer): string; override;
        function GetVerbCount: Integer; override;
        procedure Edit; override;
    end;

constructor TOCIStatementDataSetEditor.Create(AComponent: TComponent; ADesigner:
    {$IFDEF OCI_D4} IFormDesigner {$ELSE} TFormDesigner {$ENDIF});
begin
    inherited Create(AComponent, ADesigner);
    FPrevEditor := DataSetEditorClass.Create(AComponent, ADesigner);
end;

destructor TOCIStatementDataSetEditor.Destroy;
begin
    FPrevEditor.Free;
    inherited Destroy;
end;

procedure TOCIStatementDataSetEditor.ExecuteVerb(Index: Integer);
begin
    if Index < FPrevEditor.GetVerbCount then
        FPrevEditor.ExecuteVerb(Index)
    else begin
        Index := Index - FPrevEditor.GetVerbCount;
        case Index of
        0: TOCIQuery(Component).ExecSQL;
        end;
    end;
end;

function TOCIStatementDataSetEditor.GetVerb(Index: Integer): string;
begin
    if Index < FPrevEditor.GetVerbCount then
        Result := FPrevEditor.GetVerb(Index)
    else begin
        Index := Index - FPrevEditor.GetVerbCount;
        case Index of
        0: Result := '&Execute';
        end;
    end;
end;

function TOCIStatementDataSetEditor.GetVerbCount: Integer;
begin
    Result := FPrevEditor.GetVerbCount + 1;
end;

procedure TOCIStatementDataSetEditor.Edit;
begin
    ExecuteVerb(0);
end;

{ --------------------------------------------------------------------------- }
{ --------------------------------------------------------------------------- }

type
    TOCIParamNameProperty = class(TStringProperty)
        function GetAttributes: TPropertyAttributes; override;
        procedure GetValues(Proc: TGetStrProc); override;
    end;

function TOCIParamNameProperty.GetAttributes: TPropertyAttributes;
begin
    Result := [paValueList, paSortList];
end;

procedure TOCIParamNameProperty.GetValues(Proc: TGetStrProc);
var
    i: Integer;
begin
    with GetComponent(0) as TOCINestedDataSet do
        if ParamDataSet <> nil then
            for i := 0 to ParamDataSet.ParamCount - 1 do
                Proc(ParamDataSet.Params[i].Name);
end;

{ --------------------------------------------------------------------------- }
{ --------------------------------------------------------------------------- }

{$IFDEF OCI_D5}
class function TOCILoginCategory.Name: String;
begin
    Result := SOCILoginCategoryName;
end;

class function TOCILoginCategory.Description: String;
begin
    Result := SOCILoginCategoryDesc;
end;

class function TOCIResourceCategory.Name: String;
begin
    Result := SOCIResourceCategoryName;
end;

class function TOCIResourceCategory.Description: String;
begin
    Result := SOCIResourceCategoryDesc;
end;

class function TOCITransactCategory.Name: String;
begin
    Result := SOCITransactCategoryName;
end;

class function TOCITransactCategory.Description: String;
begin
    Result := SOCITransactCategoryDesc;
end;
{$ENDIF}

{ --------------------------------------------------------------------------- }
{ --------------------------------------------------------------------------- }

procedure Register;
begin
    CaptureDataSetEditorClass;
    
    RegisterPropertyEditor(TypeInfo(String), TOCIDatabase, 'ServerName',
        TOCIServerNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCIDataSet, 'DatabaseName',
        TOCIDatabaseNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCITransactionManager, 'DatabaseName',
        TOCIDatabaseNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCIStoredProc, 'StoredProcName',
        TOCIStoredProcNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCIStoredProc, 'OPackageName',
        TOCIOPackageNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCIStoredProc, 'OProcedureName',
        TOCIOProcedureNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCISequence, 'DatabaseName',
        TOCIDatabaseNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCISequence, 'SequenceName',
        TOCISequenceNameProperty);
    RegisterPropertyEditor(TypeInfo(String), TOCINestedDataSet, 'ParamName',
        TOCIParamNameProperty);

    RegisterComponentEditor(TOCIQuery, TOCIStatementDataSetEditor);
    RegisterComponentEditor(TOCIStoredProc, TOCIStatementDataSetEditor);

{$IFDEF OCI_D5}
    RegisterPropertiesInCategory(TActionCategory, ['Enabled']);

    RegisterPropertiesInCategory(TOCITransactCategory, TOCIDatabase,
        ['AutoCommit', 'Trans*', 'BeforeCommit', 'AfterCommit',
         'BeforeRollback', 'AfterRollback', 'BeforeStartTransaction',
         'AfterStartTransaction']);
    RegisterPropertiesInCategory(TOCITransactCategory, TOCIDataSet,
        ['TransactionManager']);
    RegisterPropertiesInCategory(TOCITransactCategory, TOCITransactionManager,
        ['*']);

    RegisterPropertiesInCategory(TDatabaseCategory, TOCIDatabase,
        ['AuthentMode', 'Connect*', 'DataBaseName', 'Default*',
         'InitModes', 'KeepConnection', 'Login*', 'NonBlockingMode',
         'Password', 'ServerName', 'UserName', 'SilentMode',
         'After*', 'Before*', 'On*', 'CachedUpdates']);
    RegisterPropertiesInCategory(TDatabaseCategory, TOCIDataSet,
        ['Active', 'DatabaseName', 'DataFormat', 'FetchParams',
         'After*', 'Before*', 'On*', 'UpdateMode', 'RecordCountMode']);
    RegisterPropertiesInCategory(TDatabaseCategory, TOCICustomQuery,
        ['MacroCheck', 'Macros', 'Prepared', 'Unidirectional']);
    RegisterPropertiesInCategory(TDatabaseCategory, TOCIStoredProc,
        ['StoredProcName', 'OPackageName', 'OProcedureName',
         'Overload']);
    RegisterPropertiesInCategory(TDatabaseCategory, TOCITransactionManager,
        ['DatabaseName']);
    RegisterPropertiesInCategory(TDatabaseCategory, TOCISequence,
        ['SequenceName']);

    RegisterPropertiesInCategory(TVisualCategory, TOCIDatabase,
        ['Login*', 'SilentMode', 'Show*']);

    RegisterPropertiesInCategory(TInputCategory, TOCIDataSet,
        ['AllowedOperations']);

    RegisterPropertiesInCategory(TOCILoginCategory, TOCIDatabase,
        ['AuthentMode', 'Connect*', 'Login*', 'Password', 'ServerName',
         'UserName', 'OnLogin', 'OnChangePassword']);

    RegisterPropertiesInCategory(TOCIResourceCategory, TOCIDatabase,
        ['AutoCommit', 'Connected', 'KeepConnection', 'DefaultFetchParams',
         'InitModes', 'NonBlockingMode', '*Connect', 'Max*']);
    RegisterPropertiesInCategory(TOCIResourceCategory, TOCIDataSet,
        ['Active', 'FetchParams', '*Open', '*Close', 'Disconnectable']);
    RegisterPropertiesInCategory(TOCIResourceCategory, TOCIStatementDataSet,
        ['Prepared', '*Prepare']);
    RegisterPropertiesInCategory(TOCIResourceCategory, TOCICustomQuery,
        ['ParamCheck', 'MacroCheck']);

    RegisterPropertiesInCategory(TLinkageCategory, TOCINestedDataSet,
        ['ParamName']);
{$ENDIF}

    RegisterComponents('Data Access', [TOCIDatabase, TOCITransactionManager,
        TOCIQuery, TOCIStoredProc, TNCSQLMonitorClient, TOCIUpdateSQL,
        TOCISequence, TOCINestedDataSet]);
end;

end.
