{*******************************************************}
{File:      NCOciUtil.PAS                               }
{Revision:  0.01.07 / 24.08.2000                        }
{Comment:   NC OCI8 VCL: utility unit                   }
{Copyright: (c) 1999, 2000, Dmitry Arefiev              }
{Author:    Dmitry Arefiev, dmitrya@inthink.com         }
{*******************************************************}
{$I NCOciDef.inc}

unit NCOciUtil;

interface

{$IFDEF OCI_D4}
function StrLike(const AStr, AMask: String; ANoCase: Boolean = False;
    AManyCharsMask: Char = '%'; AOneCharMask: Char = '_'; AEscapeChar: Char = '\'): Boolean;
{$ELSE}
function StrLike(const AStr, AMask: String; ANoCase: Boolean;
    AManyCharsMask: Char; AOneCharMask: Char; AEscapeChar: Char): Boolean;
{$ENDIF}
function StrRPos(Str1, Str2: PChar): PChar;
function VerStr2Int(const AVersion: String): Integer;
function VarNearEqual(const V1, V2: Variant): Boolean;
function UCOraName(const AName: String): String;
function NormOraName(const AName: String): String;

implementation

Uses SysUtils;

{$IFDEF OCI_D4}
function StrLike(const AStr, AMask: String; ANoCase: Boolean = False;
    AManyCharsMask: Char = '%'; AOneCharMask: Char = '_'; AEscapeChar: Char = '\'): Boolean;
{$ELSE}
function StrLike(const AStr, AMask: String; ANoCase: Boolean;
    AManyCharsMask: Char; AOneCharMask: Char; AEscapeChar: Char): Boolean;
{$ENDIF}

    function InternalScan(pStr, pMask: PChar): Boolean;
    begin
        Result := True;
        while ((pStr^ = pMask^) and (pMask^ <> AManyCharsMask) or
               (pMask^ = AOneCharMask) or (pMask^ = AEscapeChar)) and
              (pStr^ <> #0) and (pMask^ <> #0) do begin
            if pMask^ = AEscapeChar then begin
                Inc(pMask);
                if pStr^ <> pMask^ then
                    Break;
            end;
            Inc(pMask);
            Inc(pStr);
        end;
        if pMask^ = AManyCharsMask then begin
            while (pMask^ = AManyCharsMask) or (pMask^ = AOneCharMask) do begin
                if pMask^ = AOneCharMask then
                    if pStr^ = #0 then begin
                        Result := False;
                        Exit;
                    end
                    else
                        Inc(pStr);
                Inc(pMask);
            end;
            if pMask^ <> #0 then begin
                while (pStr^ <> #0) and ((pMask^ <> pStr^) or not InternalScan(pStr, pMask)) do
                    Inc(pStr);
                Result := pStr^ <> #0;
            end;
        end
        else
            Result := (pMask^ = #0) and (pStr^ = #0);
    end;

begin
    if AMask = AManyCharsMask then
        Result := True
    else if (AMask = '') or (AStr = '') then
        Result := False
    else begin
        if ANoCase then
            Result := InternalScan(PChar(AnsiUpperCase(AStr)), PChar(AnsiUpperCase(AMask)))
        else
            Result := InternalScan(PChar(AStr), PChar(AMask));
    end;
end;

function StrRPos(Str1, Str2: PChar): PChar;
var
    p1, p2, pBSub, pESub, pBS, pES: PChar;
begin
    Result := nil;
    if (Str1 = nil) or (Str2 = nil) then
        Exit;
    pBSub := Str2;
    pESub := Str2 + StrLen(Str2) - 1;
    pBS := Str1 + StrLen(Str2) - 1;
    pES := Str1 + StrLen(Str1) - 1;
    while pES >= pBS do begin
        if pES^ = pESub^ then begin
            p1 := pES;
            p2 := pESub;
            while (p2 >= pBSub) and (p1^ = p2^) do begin
                Dec(p1);
                Dec(p2);
            end;
            if p2 < pBSub then begin
                Result := p1 + 1;
                Exit;
            end;
        end;
        Dec(pES);
    end;
end;

function VerStr2Int(const AVersion: String): Integer;
var
    i, i1: Integer;
    s: String;
begin
    Result := 0;
    s := AVersion;
    for i := 1 to 5 do begin
        if s = '' then
            Result := Result * 100
        else begin
            i1 := Pos('.', s);
            if i1 = 0 then
                i1 := Length(s) + 1;
            Result := Result * 100 + StrToInt(Trim(Copy(s, 1, i1 - 1)));
            s := Copy(s, i1 + 1, Length(s));
        end;            
    end;
end;

function VarNearEqual(const V1, V2: Variant): Boolean;
var
    vt1, vt2: Integer;
    dt1, dt2: TDateTime;
    d1, d2: Double;
begin
    vt1 := VarType(V1) and varTypeMask;
    vt2 := VarType(V2) and varTypeMask;
    if vt1 = vt2 then
        if vt1 = varDate then begin
            dt1 := V1;
            dt2 := V2;
            Result := Abs(dt1 - dt2) <= 1.0 / 86400.0;
        end
        else if vt1 in [varSingle, varDouble] then begin
            d1 := V1;
            d2 := V2;
            Result := Abs(d1 - d2) <= 1e-10;
        end
        else if vt1 in [varEmpty, varNull] then
            Result := True
        else
            Result := V1 = V2
    else
        Result := False;
end;

function UCOraName(const AName: String): String;
begin
    Result := TrimLeft(TrimRight(AName));
    if Result <> '' then
        if (Result[1] = '"') and (Result[Length(Result)] = '"') then
            Result := Copy(Result, 2, Length(Result) - 2)
        else
            Result := AnsiUpperCase(Result);
end;

function NormOraName(const AName: String): String;
var
    i: Integer;
begin
    Result := AName;
    for i := 1 to Length(AName) do
        if (i = 1) and not (AName[1] in ['a' .. 'z', 'A' .. 'Z']) or
           (i > 1) and not (AName[i] in ['a' .. 'z', 'A' .. 'Z',
                                         '0' .. '9', '#', '$', '_']) then begin
            Result := '"' + AName + '"';
            Break;
        end;
end;

end.
